from superhelp.helpers import all_blocks_help
from .. import ast_funcs, conf
from superhelp.gen_utils import layout_comment as layout

def imported_re(block_el):
    ## straight import
    import_els = block_el.xpath('descendant-or-self::Import/names/alias')
    import_names = [el.get('name') for el in import_els]
    if 're' in import_names:
        return True
    import_from_module_els = block_el.xpath('descendant-or-self::ImportFrom')
    from_import_names = [el.get('module') for el in import_from_module_els]
    if 're' in from_import_names:
        return True
    return False

def _imported_verbose_constant_attr(block_el):
    attr_els = block_el.xpath('descendant-or-self::value/Attribute')
    verbose_constant = False
    for attr_el in attr_els:
        if attr_el.get('attr') == conf.VERBOSE_FLAG:
            verbose_constant = True
            break
    return verbose_constant

def _from_imported_verbose_constant(block_el):
    import_from_els = block_el.xpath('descendant-or-self::ImportFrom')
    verbose_constant = False
    for import_from_el in import_from_els:
        name_alias_els = import_from_el.xpath('descendant-or-self::names/alias')
        for name_alias_el in name_alias_els:
            if name_alias_el.get('name') == conf.VERBOSE_FLAG:
                verbose_constant = True
                break
    return verbose_constant
    
def used_verbose_constant(block_el):
    if _imported_verbose_constant_attr(block_el):
        return True
    if _from_imported_verbose_constant(block_el):
        return True
    return False

def used_inline_verbose(block_el):
    used_inline_verbose = False
    ## Treating presence of '(?x)' as sign of in-line verbose mode indicator.
    ## To be honest, if it actually is there for another reason the worst that
    ## happens is we don't let them know about verbose option when we could have
    str_els = ast_funcs.str_els_from_block(block_el)
    for str_el in str_els:
        str_val = ast_funcs.str_from_el(str_el)
        if str_val and str_val.startswith(conf.INLINE_RE_VERBOSE_FLAG):
            used_inline_verbose = True
            break
    return used_inline_verbose

def used_verbose(block_el):
    if used_verbose_constant(block_el):
        return True
    if used_inline_verbose(block_el):
        return True
    return False

def used_compile(block_el):
    pass

@all_blocks_help()
def verbose_option(blocks_dets):
    """
    Check for use of regex without verbose mode and introduce the idea.
    """
    has_imported_re = False
    has_used_verbose = False
    for block_dets in blocks_dets:
        block_el = block_dets.element
        if not has_imported_re:
            if imported_re(block_el):
                has_imported_re = True
        if not has_used_verbose:
            if used_verbose(block_el):
                has_used_verbose = True
        if has_imported_re and has_used_verbose:
            break
    needs_verbose = has_imported_re and not has_used_verbose
    if not needs_verbose:
        return None

    title = layout("""\
    ### Option of using verbose mode with regex
    """)
    tm = '\N{TRADE MARK SIGN}'
    brief_explain = layout(f"""\

    Regular expressions are infamous for being hard to understand, hard to
    debug, and hard to maintain / extend. It is sometimes said that when you
    have a problem, and you solve it with regex, now you have two problems ;-).
    So anything you can do to make your regex more readable is a very Good
    Thing{tm}. Which is where verbose mode is often helpful.
    """)
    longer_explain = (
            layout("""\

            Verbose mode lets you split your regex into smaller, more manageable
            parts, and you can comment your intentions for each part. The
            following example is hard enough to understand _with_ comments -
            imagine trying to make sense of it without! Or make modifications
            without breaking everything. Good luck with that!
            """)
            +
            layout('''\
            pattern = r"""
                (?P<before_id>
                var\sdiv_icon_                    ## var div_icon
                \w+                               ## lots of characters (don't care which until ...)
                \s=\sL.divIcon\(\{[\s\S]*?\<div)  ##  = L.divIcon({ anything till <div (non-greedy)
                (?P<after_id>[\s\S]*? ;)          ## anything until first semi-colon (we're non-greedy
                                                  ## so we can pick up multiple instances of pattern)
                """
            ''', is_code=True)
            +
            layout("""\

            Note - because whitespace is ignored it must be explicitly handled
            in your pattern. In some cases the overhead of doing this will
            outweigh the benefits of using verbose mode.
            """)
            +
            layout("""\
            You can either use the verbose flag e.g.
            """)
            +
            layout("""\
            re.match(r'car', 'cardboard', flags=re.VERBOSE)
            """, is_code=True)
            +
            layout("""\
            or the in-line version of the flag e.g.
            """)
            +
            layout("""\
            re.match(r'(?x)car', 'cardboard')
            """, is_code=True)
        )
    extra = layout("""\

    The standard documentation at <https://docs.python.org/3/howto/regex.html>
    is really good. Having said that, regex is never trivial so don't worry if
    you find it challenging.

    Pronunciation: reggex or rejex? Regex with a hard 'g' probably makes most
    sense because regular also has a hard 'g' but there is no consensus on
    pronunciation.
    """)

    message = {
        conf.BRIEF: title + brief_explain,
        conf.MAIN: title + longer_explain,
        conf.EXTRA: extra,
    }
    return message


from superhelp.helpers import any_block_help, filt_block_help
from ..ast_funcs import assigned_str_els_from_block, get_str_els_being_combined
from .. import code_execution, conf, name_utils
from superhelp.gen_utils import get_nice_str_list, layout_comment as layout

F_STR = 'f-string'
STR_FORMAT_FUNC = 'str_format'
SPRINTF = 'sprintf'
STR_ADDITION = 'string addition'

ASSIGN_VALUE_XPATH = 'descendant-or-self::Assign/value'
FUNC_ATTR_XPATH = 'descendant-or-self::value/Call/func/Attribute'
JOINED_STR_XPATH = 'descendant-or-self::Assign/value/JoinedStr'
SPRINTF_XPATH = 'descendant-or-self::value/BinOp/op/Mod'
STR_ADDITION_XPATH = 'descendant-or-self::BinOp/left/Str'  ## each left has a right as well so only need to look at one to get at binop and assign ancestors etc

F_STR_REMINDER = False

@filt_block_help(xpath=ASSIGN_VALUE_XPATH)
def assigned_str_overview(block_dets, *, repeat=False):
    """
    Provide overview of assigned strings e.g. name = 'Hamish'.
    """
    str_els = assigned_str_els_from_block(block_dets.element)
    if not str_els:
        return None
    names = []
    for str_el in str_els:
        name_dets = name_utils.get_assigned_name(str_el)
        names.append(name_dets)
    name_strs = [name_dets.name_str for name_dets in names]
    multiple = (len(set(name_strs)) > 1)

    title = layout("""\
    #### String Overview
    """)
    if multiple:
        nice_list_str = get_nice_str_list(sorted(name_strs), quoter='`')
        summary = layout(f"""\
        {nice_list_str} are all strings
        """)
    else:
        summary = layout(f"""\
        `{names[0].name_str}` is a string.
        """)
    if not repeat:
        cool = layout("""\
        Python makes it easy to do lots of cool things with strings.
        """)
        first_name_dets = names[0]
        try:
            first_val = code_execution.get_val(
                block_dets.pre_block_code_str, block_dets.block_code_str,
                first_name_dets.name_type,
                first_name_dets.name_details, first_name_dets.name_str)
        except KeyError:
            name2use = 'address'
            val2use = 'Waiuku, New Zealand'
        else:
            name2use = first_name_dets.name_str
            val2use = first_val
        short_demo = layout(f"""\

        For illustration, imagine we have string '{val2use}' assigned to
        `{name2use}`:

        """)
        upper = layout(f"""\
        `{name2use}.upper()` returns {val2use.upper()}.
        """)
        black_heart = "\N{BLACK HEART}"
        longer_demo = layout(f"""\
        Examples:

        `{name2use}.upper()` returns {val2use.upper()}.

        `{name2use}.center(70, '=')` returns {val2use.center(70, '=')}

        `{name2use}.endswith('chicken')` returns {val2use.endswith('chicken')}

        `{name2use}` + ' is a string' returns {val2use + ' is a string'}

        `{name2use}` + ' ' + '{{\\NBLACK HEART}}' + ' Python' returns
        {val2use + ' ' + black_heart + ' Python'}

        `len({name2use})` returns {len(val2use)} because that is how many
        characters are in the `{name2use}` string (remember to count spaces -
        they are characters too)

        `sorted({name2use})` returns {sorted(val2use)}
        """)
        more = layout("""\

        `.upper()`, `.center()` etc are abilities available with all Python
        strings. Technically they are methods of string objects. They start with
        a dot and are on the end of the object.

        To see the full list of string methods enter `dir(str)` into a Python
        command line.

        `len()` is a function which can be used on lots of things - not just
        string objects. It is not a method of the string object. Other functions
        that are not string-specific but are commonly used with strings include
        `sorted()` and `print()`.
        """)
    else:
        cool = ''
        short_demo = ''
        upper = ''
        longer_demo = ''
        more = ''

    message = {
        conf.BRIEF: title + summary + cool + short_demo + upper,
        conf.MAIN: title + summary + cool + short_demo + upper + longer_demo,
        conf.EXTRA: more,
    }
    return message

def str_combination(combination_type, str_els, *, repeat=False):
    global F_STR_REMINDER
    combination_type2comment = {
        F_STR: "f-string interpolation",
        STR_FORMAT_FUNC: "the format function",
        SPRINTF: "sprintf string interpolation",
        STR_ADDITION: "string addition (e.g. animal = 'jelly' + 'fish')",
    }
    title = layout("""\
    ### Strings created by combining or interpolating strings
    """)
    how_combined_bits = []
    for str_el in str_els:
        try:
            assign_el = str_el.xpath('ancestor::Assign')[-1]
        except IndexError:
            name = "An unnamed string"
        else:
            raw_name = assign_el.xpath('targets/Name')[0].get('id')
            name = f"`{raw_name}`"
        combination_comment = combination_type2comment[combination_type]
        how_combined_bits.append(layout(f"""\

        `{name}` is created using {combination_comment}.
        """))
    how_combined = ''.join(how_combined_bits)
    if not repeat:
        if combination_type != F_STR and not F_STR_REMINDER:
            F_STR_REMINDER = True
            brief_fstring_msg = layout("""\

            Your snippet uses a non-f-string approach to constructing a string.

            Have you considered using an f-string approach to constructing your
            string?
            """)
            longer_fstring_msg = (
                layout("""\

                Have you considered using an f-string approach to constructing
                your string?

                f-strings let you reference variables from earlier in your code
                and allow very readable string construction. All the usual
                tricks of the `.format()` approach also work. For example, comma
                separating thousands in numbers:
                """)
                +
                layout("""\
                cost = 10_550
                print(f"Cost is ${cost:,}")
                # >>> 'Cost is $10,550'
                """, is_code=True)
                +
                layout("""\
                Or making small in-line calculations:
                """)
                +
                layout("""\
                people = ['Bart', 'Lisa']
                print(f"There are {len(people)} people")
                # >>> 'There are 2 people'
                """, is_code=True)
                +
                layout("""\
                Or zero-padding numbers:
                """)
                +
                layout("""\
                num = 525
                print(f"{num:0>4}")
                # >>> '0525'
                """, is_code=True)
            )
        else:
            brief_fstring_msg = ''
            longer_fstring_msg = ''
    else:
        brief_fstring_msg = ''
        longer_fstring_msg = ''

    message = {
        conf.BRIEF: title + how_combined + brief_fstring_msg,
        conf.MAIN: title + how_combined + longer_fstring_msg,
    }
    return message

@filt_block_help(xpath=JOINED_STR_XPATH)
def f_str_interpolation(block_dets, *, repeat=False):
    """
    Examine f-string interpolation.
    """
    joined_els = block_dets.element.xpath(JOINED_STR_XPATH)
    return str_combination(F_STR,
        joined_els, repeat=repeat)

@filt_block_help(xpath=FUNC_ATTR_XPATH)
def format_str_interpolation(block_dets, repeat=False):
    """
    Look at use of .format() to interpolate into strings.
    """
    func_attr_els = block_dets.element.xpath(FUNC_ATTR_XPATH)
    format_funcs = []
    for func_attr_el in func_attr_els:
        is_format_func = func_attr_el.get('attr') == 'format'
        if is_format_func:
            format_funcs.append(func_attr_el)
    if not format_funcs:
        return None
    return str_combination(STR_FORMAT_FUNC,
        format_funcs, repeat=repeat)

@any_block_help()
def sprintf(block_dets, *, repeat=False):
    """
    Look at use of sprintf for string interpolation
    e.g. greeting = "Hi %s" % name
    """
    sprintf_els = block_dets.element.xpath(SPRINTF_XPATH)
    has_sprintf = bool(sprintf_els)
    if not has_sprintf:
        return None
    return str_combination(SPRINTF,
        sprintf_els, repeat=repeat)

@any_block_help()
def string_addition(block_dets, *, repeat=False):
    """
    Advise on string combination using +.
    Explain how f-string alternative works.
    """
    str_els_being_combined = get_str_els_being_combined(block_dets.element)
    has_string_addition = False
    str_addition_els = []
    for str_el in str_els_being_combined:
        ## Does their immediate ancestor BinOp have op of Add?
        ## Don't know if there any alternatives but let's be sure
        ## Ordered set of nodes, from parent to ancestor? https://stackoverflow.com/a/15645846
        bin_op_el = str_el.xpath('ancestor-or-self::BinOp')[-1]
        ## was it an Add op
        has_add = bool(bin_op_el.xpath('op/Add'))
        if has_add:
            str_addition_els.append(str_el)
            has_string_addition = True
    if not has_string_addition:
        return None
    addition_message = str_combination(STR_ADDITION,
        str_addition_els, repeat=repeat)
    return addition_message



from html import unescape
from markdown.extensions import Extension  # @UnresolvedImport
from markdown.treeprocessors import Treeprocessor  # @UnresolvedImport
from tabulate import tabulate  # @UnresolvedImport

from . import cli_colour, cli_conf, cli_utils, tag_formatting


class AnsiPrinter(Treeprocessor):

    @staticmethod
    def handle_admonitions(text):
        """
        Admonitions are blocks that stand out and might start with an icon e.g.
        alert. They are indicated by a leading '!!! ' and have other words with
        space separation
        """
        admon_lbl_used = None
        prefix = body_prefix = ''
        len_admon_start = len(cli_conf.ADMON_START)
        if text.startswith(cli_conf.ADMON_START):
            ## we have to handle admon labels with spaces so check with startswith not ==
            for admon_lbl in cli_colour.ADMONS:
                if text[len_admon_start:].startswith(admon_lbl):
                    break
            else:  ## i.e. never exited
                admon_lbl = text[len_admon_start:].split(" ", 1)[0]
                cli_colour.ADMONS[admon_lbl] = cli_colour.ADMONS.values()[0]
            prefix = body_prefix = '┃ '
            prefix += admon_lbl.capitalize()
            admon_lbl_used = admon_lbl
            text = text.split(admon_lbl, 1)[1]
        return text, prefix, body_prefix, admon_lbl_used 

    @staticmethod
    def handle_blockquotes(el, out, *, nesting_level):
        for el1 in el.getchildren():
            iout = []
            AnsiPrinter.formatter(el1, iout, nesting_level + 2, parent=el)
            pr = cli_colour.colourise(
                cli_conf.BQUOTE_PREFIX, cli_colour.H1_COLOUR)
            sp = ' ' * (nesting_level + 2)
            for l in iout:
                for l1 in l.splitlines():
                    if sp in l1:
                        l1 = ''.join(l1.split(sp, 1))
                    out.append(pr + l1)

    @staticmethod
    def handle_list_item(el, out, *, nesting_level):
        childs = el.getchildren()
        for nested in ['ul', 'ol']:
            if childs and childs[-1].tag == nested:
                ul = childs[-1]
                # Nested sublist? The li was inline formatted so
                # split all from <ul> off and format it as own tag:
                # (ul always at the end of an li)
                out[-1] = out[-1].split(f"<{nested}>", 1)[0]
                AnsiPrinter.formatter(
                    ul, out, nesting_level + 1, parent=el)

    @staticmethod
    def handle_table(el, out, *, nesting_level):
        """
        All processed here in one sweep. Markdown ext gave us a xml tree from
        the ascii. Our part here is the cell formatting and into a python nested
        list, then tabulate spits out ascii again.
        """
        def fmt(cell, parent):
            """ we just run the whole formatter - just with a fresh new
            result list so that our 'out' is untouched """
            _cell = []
            AnsiPrinter.formatter(
                cell, out=_cell, nesting_level=0, parent=parent)
            return "\n".join(_cell)

        text = []
        for el_idx in 0, 1:
            for Row in el[el_idx].getchildren():
                row = []
                text.append(row)
                for cell in Row.getchildren():
                    row.append(fmt(cell, row))
        # good ansi handling:
        tbl = tabulate(text)
        # do we have right room to indent it?
        # first line is seps, so no ansi escapes foo:
        width = len(tbl.split('\n', 1)[0])
        if width <= cli_conf.TERMINAL_WIDTH:
            text = tbl.splitlines()
            cli_utils.apply_borders(text)
            # center:
            indent = (cli_conf.TERMINAL_WIDTH - width) / 2
            # too much:
            indent = nesting_level
            tt = []
            for line in text:
                tot_indent = indent * cli_conf.LEFT_INDENT
                tt.append(f"{tot_indent}{line}")
            out.extend(tt)
        else:
            # TABLE CUTTING WHEN NOT WIDTH FIT
            # oh snap, the table bigger than our screen. hmm.
            # hey lets split into vertical parts:
            # but len calcs are hard, since we are crammed with escaping
            # seqs.
            # -> get rid of them:
            tc = []
            for row in text:
                tc.append([])
                l = tc[-1]
                for cell in row:
                    l.append(cli_utils.clean_ansi(cell))
            # again sam:
            # note: we had to patch it, it inserted '\n' within cells!
            table = tabulate(tc)
            out.append(
                cli_utils.split_blocks(
                    text_block=table, width=width,
                    n_cols=cli_conf.TERMINAL_WIDTH,
                    part_formatter=cli_utils.apply_borders)
            )

    @staticmethod
    def handle_lists(el, out, *, nesting_level):
        nr = 0
        for c in el:
            if el.tag == 'ul':
                c.set('prefix', cli_conf.LIST_PREFIX)
            elif el.tag == 'ol':
                nr += 1
                c.set('prefix', str(nr) + '. ')
            # handle the ``` style unindented code blocks -> parsed as p:
            AnsiPrinter.formatter(c, out, nesting_level + 1, parent=el)

    @staticmethod
    def formatter(el, out, nesting_level=0, prefix='', parent=None):
        """
        Main recursion down tree applying ANSI formatting.

        :param int nesting_level: nesting_level
        :param str prefix: might be set by admonitions e.g. '| WARNING' or set
         by parent ol ('1.', '2.' etc) or ul ('-'). Lists seems to override
         admonitions if the admonition has internal lists. A bug?
        """
        if el.tag == 'br':
            out.append('\n')
            return

        links_list = None
        is_txt_and_inline_markup = False

        if el.tag == 'blockquote':
            AnsiPrinter.handle_blockquotes(el, out, nesting_level=nesting_level)
            return

        if el.tag == 'hr':
            out.append(tag_formatting.hr('', nesting_level=nesting_level))
            return

        is_text = (
            bool(el.text)
            or el.tag == 'p'
            or el.tag == 'li'
            or el.tag.startswith('h')
        )
        if is_text:
            el.text = el.text or ''
            # <a attributes>foo... -> we want "foo....". Is it a sub
            # tag or inline text?
            text_with_inline_markup = cli_utils.get_text_if_inline_markup(el)
            if text_with_inline_markup:
                is_txt_and_inline_markup = True
                # foo:  \nbar -> will be seeing a foo:<br>bar with
                # mardkown.py. Code blocks are already quoted -> no prob.
                text_with_inline_markup = text_with_inline_markup.replace(
                    '<br />', '\n')
                # strip our own closing tag:
                text = text_with_inline_markup.rsplit('<', 1)[0]
                links_list, text = cli_utils.replace_links(
                    el, html=text, link_display_type='it')
                for tg, (start, end) in cli_conf.TAG2BOUNDS.items():
                    text = text.replace(tg, start)
                    tag_lbl = tg[1:]
                    close_tag = f"</{tag_lbl}"
                    text = text.replace(close_tag, end)
                text = unescape(text)
            else:
                text = el.text
            if cli_conf.BADLY_PARSED_UNDERSCORE in text:
                text = text.replace(cli_conf.BADLY_PARSED_UNDERSCORE, '_')  ## so __doc__ is displayed not 9595doc9595 when the source md has \_\_doc|_|_
            text = text.strip()

            admon_res = AnsiPrinter.handle_admonitions(text)
            text, prefix, body_prefix, admon_lbl_used = admon_res
            # set the parent, e.g. nrs in ols:
            if el.get('prefix'):
                # first line prefix, like '-':
                prefix = el.get('prefix')
                # next line prefs:
                body_prefix = ' ' * len(prefix)
                el.set('prefix', '')

            indent = cli_conf.LEFT_INDENT * nesting_level
            if el.tag in cli_conf.HEADER_TAGS:
                header_level = int(el.tag[1:])
                indent = ' ' # * (header_level - 1)
                nesting_level += header_level

            text = cli_utils.rewrap(el, text, indent, prefix,
                terminal_width=cli_conf.TERMINAL_WIDTH)

            # indent. can color the prefixes now, no more len checks:
            if admon_lbl_used:
                out.append("\n")
                colour = globals()[cli_colour.ADMONS[admon_lbl_used]]
                prefix = cli_colour.colourise(prefix, colour)
                body_prefix = cli_colour.colourise(body_prefix, colour)

            if prefix:
                h_level = ((nesting_level - 2) % 5) + 1
                colour = getattr(cli_colour, f"H{h_level}_COLOUR")
                if (prefix == cli_conf.LIST_PREFIX
                        or prefix.split('.', 1)[0].isdigit()):
                    prefix = cli_colour.colourise(prefix, colour)

            text = ('\n' + indent + body_prefix).join((text).splitlines())
            text = indent + prefix + text

            # headers outer left: go sure.
            # actually... NO. commented out.
            # if el.tag in self.header_tags:
            #    prefix = ''

            # calling the class Tags  functions
            # IF the parent is li and we have a linebreak then the renderer
            # delivers <li><p>foo</p> instead of <li>foo, i.e. we have to
            # omit the linebreak and append the text of p to the previous
            # result, (i.e. the list separator):
            tag_fmt_func = getattr(tag_formatting, el.tag,
                cli_colour.colourise_plain)  ## default
            if (
                type(parent) == type(el)
                and parent.tag == 'li'
                and not parent.text
                and el.tag == 'p'
            ):
                _out = tag_fmt_func(
                    text.lstrip(), nesting_level=nesting_level)
                out[-1] += _out
            else:
                out.append(
                    tag_fmt_func(text, nesting_level=nesting_level))

            if admon_lbl_used:
                out.append('\n')

            if links_list:
                for i, l in enumerate(links_list, 1):
                    out.append(cli_colour.colourise_low_vis(
                        f"{indent}[{i}] {l}"))

        if is_txt_and_inline_markup:
            if el.tag == 'li':
                AnsiPrinter.handle_list_item(
                    el, out, nesting_level=nesting_level)
            return

        if el.tag == 'table':
            AnsiPrinter.handle_table(el, out, nesting_level=nesting_level)
            return

        AnsiPrinter.handle_lists(el, out, nesting_level=nesting_level)

    def run(self, doc):
        out = []
        AnsiPrinter.formatter(doc, out)
        ansi = '\n'.join(out)
        self.markdown.ansi = ansi


class AnsiPrintExtension(Extension):
    def extendMarkdown(self, md, md_globals):
        ansi_print_ext = AnsiPrinter(md)
        md.treeprocessors.add('ansi_print_ext', ansi_print_ext, '>inline')



from superhelp.helpers import filt_block_help
from .. import code_execution, conf
from superhelp import gen_utils
from superhelp.gen_utils import layout_comment as layout

ASSIGN_LIST_XPATH = 'descendant-or-self::Assign/value/List'

DEFAULT_EXAMPLE_TYPE = conf.STR_TYPE

def get_item_type_names(items):
    item_type_names = sorted(set(
        [type(item).__name__ for item in items]
    ))
    item_type_nice_names = [
        conf.TYPE2NAME.get(item_type, item_type)
        for item_type in item_type_names]
    return item_type_names, item_type_nice_names

def get_type_for_example(list_items):
    """
    If all else fails, just use strings for examples.
    """
    if list_items is None:
        type4example = DEFAULT_EXAMPLE_TYPE
    else:
        item_type_names, _item_type_nice_names = get_item_type_names(list_items)
        try:
            type4example = item_type_names.pop()
        except IndexError:
            type4example = DEFAULT_EXAMPLE_TYPE
    return type4example

def _get_detailed_list_comment(first_name, first_items):
    type4example = get_type_for_example(first_items)
    try:
        example_items = conf.EXAMPLES_OF_TYPES[type4example]
    except KeyError:
        example_items = conf.EXAMPLES_OF_TYPES[DEFAULT_EXAMPLE_TYPE]
    example_item = example_items[0]
    listable_example_item = (f"'{example_item}'"
        if type4example == conf.STR_TYPE else example_item)
    if first_items is None:
        appended_list = []
    else:
        appended_list = first_items.copy()
    appended_list.append(example_item)
    extended_list = appended_list.copy()
    items2extend = example_items[1:3]  ## don't repeat the appended item - might confuse the user at this stage
    extended_list.extend(items2extend)
    friends = ['Selma', 'Willy', 'Principal Skinner']
    family = ['Bart', 'Lisa', 'Marge', 'Homer']
    guests = friends + family
    detailed_list_comment = (
        layout("""\

        Lists, along with dictionaries, are the workhorses of Python data
        structures.

        Lists have an order, and can contain duplicate items and items of
        different types (usually not advisable).

        Extra items can be added to lists using the .append() method e.g.
        """)
        +
        layout(f"""\
        {first_name}.append({listable_example_item})
        """, is_code=True)
        +
        layout(f"""\
        which results in {appended_list}
        """)
        +
        layout("""\

        If you want to add multiple items at once, .extend() is useful.

        Note - we are extending the list which has already had an item appended,
        not the original list.
        """)
        +
        layout(f"""\
        {first_name}.extend({items2extend})
        """, is_code=True)
        +
        layout(f"""\
        which results in {extended_list}
        """)
        +
        layout("""\

        GOTCHA: if you are adding tuples to your list it is easy to forget the
        nested parentheses. E.g.
        """)
        +
        layout(f"""\
        coordinates.append((x, y))  ## Correct

        coordinates.append(x, y)  ## Oops - append only takes one item not two
        """, is_code=True)
        +
        layout("""\
        Lists can also be added together e.g.
        """)
        +
        layout(f"""\
        friends = {friends}
        family = {family}
        guests = friends + family
        """, is_code=True)
        +
        layout(f"""\
        resulting in {guests}

        GOTCHA: you can't add individual items to lists unless you put them in a
        list as well E.g.
        """)
        +
        layout(f"""\
        workmate = 'Carl'
        guests = {friends} + {family} + workmate  ## Oops - can only add lists
        guests = {friends} + {family} + [workmate]  ## That's better
        """, is_code=True)
        +
        layout(f"""\
        resulting in {guests + ['Carl']}
        """)
        )
    return detailed_list_comment

def truncate_list(items):
    return items[: conf.MAX_ITEMS_EVALUATED]

## only interested in lists when being assigned as a value
## (i.e. <body><Assign><value><List> so we're looking for List under value only)
@filt_block_help(xpath=ASSIGN_LIST_XPATH)
def list_overview(block_dets, *, repeat=False):
    """
    General overview of list taking content details into account.
    """
    list_els = block_dets.element.xpath(ASSIGN_LIST_XPATH)
    plural = 's' if len(list_els) > 1 else ''
    title = layout(f"""\
    ### List{plural} defined
    """)
    first_name = None
    first_items = None
    names_items, oversized_msg = code_execution.get_collections_dets(
        list_els, block_dets,
        collection_plural='lists', truncated_items_func=truncate_list)
    for name, items in names_items:
        if items is None:
            list_desc = layout(f"""\
            `{name}` is a list. Unable to evaluate items.
            """)
        else:
            if not first_name:
                first_name = name
                first_items = items
            if items == []:
                list_desc = layout(f"""\
                `{name}` is an empty list.
                """)
            else:
                list_desc = layout(f"""\

                `{name}` is a list with {gen_utils.int2nice(len(items))} items.
                """)
    if not first_name:
        first_name = 'demo_list'
        first_items = ['apple', 'banana', 'cherry', ]
    if not repeat:
        brief_overview = layout("""\

        Lists, along with dictionaries, are the workhorses of Python data
        structures.

        Lists have an order, and can contain duplicate items and items of
        different types (usually not advisable).
        """)
        detailed_list_comment = _get_detailed_list_comment(
            first_name, first_items)
    else:
        brief_overview = ''
        detailed_list_comment = ''

    message = {
        conf.BRIEF: title + oversized_msg + list_desc + brief_overview,
        conf.MAIN: title + oversized_msg + list_desc + detailed_list_comment,
    }
    return message

@filt_block_help(xpath=ASSIGN_LIST_XPATH, warning=True)
def mixed_list_types(block_dets, *, repeat=False):  # @UnusedVariable
    """
    Warns about lists containing a mix of data types.
    """
    list_els = block_dets.element.xpath(ASSIGN_LIST_XPATH)
    list_dets = []
    names_items, oversized_msg = code_execution.get_collections_dets(
        list_els, block_dets,
        collection_plural='lists', truncated_items_func=truncate_list)
    has_mixed = False
    for name, items in names_items:
        if items is None:
            continue
        else:
            _item_type_names, item_type_nice_names = get_item_type_names(items)
            if len(item_type_nice_names) <= 1:
                ## No explanation needed if there aren't multiple types.
                continue
            has_mixed = True
            list_dets.append((name, item_type_nice_names))
    if not has_mixed:
        return None

    title = layout("""\
    ### List(s) with mix of different data types
    """)
    mixed_warning_bits = []
    for name, item_type_nice_names in list_dets:
        mixed_warning_bits.append(layout(f"""

        `{name}` contains more than one data type - which is probably a bad
        idea.
        """))
    mixed_warning = ''.join(mixed_warning_bits)
    mixed_dets = layout(f"""\

    The data types found were: {", ".join(item_type_nice_names)}.
    """)

    message = {
        conf.BRIEF: title + oversized_msg + mixed_warning,
        conf.MAIN: title + oversized_msg + mixed_warning + mixed_dets,
    }
    return message



from superhelp.helpers import filt_block_help
from .. import code_execution, conf
from superhelp import gen_utils
from superhelp.gen_utils import layout_comment as layout

def truncate_tuple(items):
    return tuple(items[: conf.MAX_ITEMS_EVALUATED])

ASSIGN_TUPLE_XPATH = 'descendant-or-self::Assign/value/Tuple'

@filt_block_help(xpath=ASSIGN_TUPLE_XPATH)
def tuple_overview(block_dets, *, repeat=False):
    """
    Explain usage of tuples.
    """
    tup_els = block_dets.element.xpath(ASSIGN_TUPLE_XPATH)
    names_items, oversized_msg = code_execution.get_collections_dets(
        tup_els, block_dets,
        collection_plural='tuples', truncated_items_func=truncate_tuple)
    assigned_names_items = [
        (name, items) for name, items in names_items if name]
    if not assigned_names_items:
        return None

    title = layout("""\
    #### Tuple Overview
    """)
    summary_bits = []
    for name, tup in assigned_names_items:
        if tup:
            summary_bits.append(layout(f"""\
            `{name}` is a tuple. Unable to evaluate items.
            """))
        else:
            summary_bits.append(layout(f"""\
            `{name}` is a tuple with {gen_utils.int2nice(len(tup))} items.
            """))
    summary = ''.join(summary_bits)
    if not repeat:
        quick_immutability = layout("""\

        Tuples are like lists but the items inside cannot be replaced, removed,
        or added to. For example, if we have a list [1, 2] we can append a 3 to
        it. But if we have a tuple (1, 2) we cannot.
        """)
        pre_immutability = layout("""\

        Tuples are like lists but they are immutable. That means unchangeable.
        """)
        why_immutability = layout("""\

        But why would we want a data structure with all those limitations -
        wouldn't a list always be better? In practice it is often useful to know
        that a data structure is not being mutated somewhere inside the program.
        This guarantee makes it easier to reason about what the program is doing
        and what it cannot be doing. We can also use named tuples to improve
        readability.
        """)
        ordered = layout("""\

        Tuples have an order, and can contain duplicate items and items of
        different types (usually not advisable).
        """)
        non_empty_name_tups = [
            (name, tup) for name, tup in assigned_names_items if tup]
        immutability_question = ("So what is tuple immutability in practice? "
            "It means that tuple items:")
        if non_empty_name_tups:
            first_name, first_tup = non_empty_name_tups[0]
            tup_replaced = list(first_tup)
            tup_replaced[0] = 100
            tup_replaced = tuple(tup_replaced)
            tup_popped = list(first_tup)
            tup_popped.pop()
            tup_popped = tuple(tup_popped)
            tup_appended = list(first_tup)
            tup_appended.append(3)
            tup_appended = tuple(tup_appended)
            longer_immutability = layout(f"""
            {immutability_question}

            * cannot be *replaced* - so we can't run `{first_name}`[0] = 100 to
            get {tup_replaced}. It will raise an exception - TypeError:
            'tuple' object does not support item assignment)

            * cannot be *removed* - so we can't run `{first_name}`.pop() to get
            {tup_popped}.

            * cannot be *added* - so we can't run `{name}`.append(3) to get
            {tup_appended}.
            """)
        else:
            longer_immutability = layout(f"""
            {immutability_question}

            * cannot be *replaced*. It will raise an exception - TypeError:
            'tuple' object does not support item assignment)

            * cannot be *removed*

            * cannot be *added*
            """)
        friends = ['Selma', 'Willy', 'Principal Skinner']
        family = ['Bart', 'Lisa', 'Marge', 'Homer']
        original_guests = (friends, family)
        guests = (friends + ['Lenny'], family)
        extra = (
            layout("""\
            #### GOTCHA - immutable means 100% unchangeable right?

            If you have a tuple of mutable data e.g. lists, the content of those
            lists can be altered even while it remains true that you can't
            replace, remove, or add items inside the tuple.

            Demonstration:
            """)
            +
            layout(f"""\
            friends = {friends}
            family = {family}
            guests = (friends, family)
            """, is_code=True)
            +
            layout(f"""\

            `guests` is now {original_guests}. And because `guests` is immutable
            we can't add, remove, or replace its items. But look what happens to
            the `guests` tuple when we append a person to friends:
            """)
            +
            layout(f"""\
            friends.append('Lenny')
            """, is_code=True)
            +
            layout(f"""\

            `guests` is now {guests}. An immutable data structure has changed!
            In reality it is the references to individual items that are
            immutable e.g. a reference to the memory address where a mutable
            list sits.

            #### Named tuples

            Tuples are a light-weight way of defining and storing data. It is
            easier, for example, to write `coord` = (-37, 174) than it is to
            write `coord` = {{'lat': -37, 'lon': 174}}. But it can be risky
            consuming data from tuples in later code if the order or number of
            items changes. What if the tuples you are processing change from
            (lat, lon) to (lon, lat)? References to coord[0] would have changed
            meaning without it being obvious. There can be a high risk of bugs.

            Named tuples retain the light-weight advantages of plain vanilla
            tuples but enable us to use a dot notation to reference internal
            items e.g. house_lat = coord.lat. Named tuples also contain
            information on their contents when printed / logged which can make
            debugging much easier.

            Example syntax:
            """)
            +
            layout(f"""\
            from collections import namedtuple
            Coord = namedtuple('Coordinate', 'x, y')
            coord = Coord(-37, 174)
            print(coord)
            # >>> Coordinate(x=-37, y=174))
            """, is_code=True)
            +
            layout(f"""\
            #### "Tupple" vs "Toople"

            There is no consensus on how to pronounce tuple. Should it rhyme
            with cup or with hoop? People from a more mathematical background
            often prefer "toople". Others prefer to follow the spelling and say
            "tupple".
            """)
        )
    else:
        quick_immutability = ''
        pre_immutability = ''
        why_immutability = ''
        ordered = ''
        longer_immutability = ''
        extra = ''

    message = {
        conf.BRIEF: (title + oversized_msg + summary + quick_immutability
            + why_immutability + ordered),
        conf.MAIN: (title + oversized_msg + summary + pre_immutability
            + why_immutability + longer_immutability + ordered),
        conf.EXTRA: extra,
    }
    return message



from collections import namedtuple
import logging

from . import ast_funcs, conf, helpers
from .gen_utils import (get_docstring_start, get_tree,
    layout_comment as layout, xml_from_tree)

BlockDets = namedtuple(
    'BlockDets', 'element, pre_block_code_str, block_code_str, first_line_no')
BlockDets.pre_block_code_str.__doc__ = ("The code up until the line we are "
    "interested in needs to be run - it may depend on names from earlier")

MessageDets = namedtuple('MessageDets',
    'code_str, message, first_line_no, warning, source')
MessageDets.__doc__ += (
    "All the bits and pieces that might be needed to craft a message")
MessageDets.code_str.__doc__ = ("The block of code the message relates to")
MessageDets.source.__doc__ = ("A unique identifier of the source of message "
    "- useful for auditing / testing")

def get_blocks_dets(snippet, snippet_block_els):
    """
    Returning a list of all the details needed to process a line
    (namely BlockDets named tuples)

    Note - lines in the XML sit immediately under body.

    :return: list of BlockDets named tuples
    :rtype: list
    """
    snippet_lines = snippet.split('\n')
    blocks_dets = []
    for snippet_block_el in snippet_block_els:
        first_line_no, last_line_no, _el_lines_n = ast_funcs.get_el_lines_dets(
            snippet_block_el)
        block_code_str = (
            '\n'.join(snippet_lines[first_line_no - 1: last_line_no]).strip())
        pre_block_code_str = (
            '\n'.join(snippet_lines[0: first_line_no - 1]).strip()
            + '\n')
        blocks_dets.append(BlockDets(
            snippet_block_el, pre_block_code_str, block_code_str, first_line_no))
    return blocks_dets

def complete_message(message, *, source):
    """
    Ensure a complete message with all levels is available.

    :param dict message: a dict with as many as three levels
     e.g. conf.BRIEF, MAIN, EXTRA.
    :param str source: usually helper name but might be the system sending the
     message e.g. if no advice given or a bug happened.
    :return: fixed version of original dict with all levels populated
    :rtype: dict
    """
    try:
        message_keys = message.keys()
    except TypeError:
        raise TypeError(f"message had no keys - message: {message} "
            f"(type {type(message).__name__})")
    if conf.BRIEF not in message_keys:
        raise Exception(f"'{source}' gave a message lacking the mandatory "
            f"'{conf.BRIEF}' level")
    if conf.MAIN not in message_keys:
        message[conf.MAIN] = message[conf.BRIEF]
    if conf.EXTRA not in message_keys:
        message[conf.EXTRA] = ''
    return message

def get_message_dets_from_input(helper_dets, *,
        helper_input, code_str, first_line_no, repeat=False):
    name = helper_dets.helper_name
    docstring = helper_dets.helper.__doc__
    if not docstring:
        raise Exception(f'Helper "{name}" lacks a docstring - add one!')
    try:
        try:
            message = helper_dets.helper(helper_input, repeat=repeat)
        except TypeError as e:
            if "unexpected keyword argument 'repeat'" in str(e):
                message = helper_dets.helper(helper_input)
            else:
                raise
    except Exception as e:
        brief_name = '.'.join(name.split('.')[-2:])  ## last two parts only
        message = {
            conf.BRIEF: (
                layout(f"""\

                    ### Helper "`{brief_name}`" unable to run

                    Helper {brief_name} unable to run. Helper description:
                    """)
                +  ## show first line of docstring (subsequent lines might have more technical, internally-oriented comments)
                layout(get_docstring_start(docstring) + '\n')
                +
                layout(str(e))
            )
        }
        source = conf.SYSTEM_MESSAGE
        warning = True
    else:
        source = name
        warning = helper_dets.warning
        if message is None:
            return None
    message = complete_message(message, source=source)
    message_dets = MessageDets(
        code_str, message, first_line_no, warning, source=source)
    return message_dets

def _get_ancestor_block_element(element):
    """
    The item immediately under body that this element is under.
    """
    ancestor_elements = element.xpath('ancestor-or-self::*')
    ancestor_block_element = ancestor_elements[2]  ## [0] will be Module, 1 is body, and blocks are the children of body
    return ancestor_block_element

def _get_filtered_blocks_dets(helper_dets, xml, blocks_dets):
    """
    Identify source block elements according to xpath supplied. Then filter
    blocks_dets accordingly.

    :return: filtered_blocks_dets
    :rtype: list
    """
    matching_elements = xml.xpath(helper_dets.xpath)
    if matching_elements:
        logging.debug(f"{helper_dets.helper_name} had at least one match")
    else:
        logging.debug(f"{helper_dets.helper_name} had no matches")
    filt_block_elements = set(
        [_get_ancestor_block_element(element) for element in matching_elements])
    filtered_blocks_dets = [block_dets for block_dets in blocks_dets
        if block_dets.element in filt_block_elements]
    if filtered_blocks_dets:
        logging.debug(f"{helper_dets.helper_name} had at least one block")
    else:
        logging.debug(f"{helper_dets.helper_name} had no blocks")
    return filtered_blocks_dets

def get_block_level_messages_dets(blocks_dets, xml, *,
        warnings_only=False, repeat_set=None):
    """
    For each helper, get advice on every relevant block. Element type specific
    helpers process filtered blocks_dets; all block helpers process all blocks
    (as you'd expect ;-)).

    As we iterate through the blocks, only the first block under an helper
    should get the full message.
    """
    messages_dets = []
    all_helpers_dets = helpers.FILT_BLOCK_HELPERS + helpers.ANY_BLOCK_HELPERS
    for helper_dets in all_helpers_dets:
        logging.debug(f"About to process '{helper_dets.helper_name}'")
        if warnings_only and not helper_dets.warning:
            continue
        element_filtering = hasattr(helper_dets, 'xpath')
        if element_filtering:
            filtered_blocks_dets = _get_filtered_blocks_dets(
                helper_dets, xml, blocks_dets)
            blocks_dets2use = filtered_blocks_dets
            logging.debug(
                f"'{helper_dets.helper_name}' has element filtering for "
                f"{len(blocks_dets2use)} matching blocks")
        else:  ## no filtering by element type so process all blocks
            blocks_dets2use = blocks_dets
        for block_dets in blocks_dets2use:
            repeat = (helper_dets.helper_name in repeat_set)
            message_dets = get_message_dets_from_input(helper_dets,
                helper_input=block_dets, code_str=block_dets.block_code_str,
                first_line_no=block_dets.first_line_no,
                repeat=repeat)
            if message_dets:
                repeat_set.add(helper_dets.helper_name)
                messages_dets.append(message_dets)
    return messages_dets

def get_overall_snippet_messages_dets(snippet, blocks_dets, *,
        warnings_only=False, repeat_set=None):
    """
    Returns messages which apply to snippet as a whole, not just specific
    blocks. E.g. looking at every block to look for opportunities to unpack. Or
    reporting on linting results.
    """
    messages_dets = []
    all_helpers_dets = (
        helpers.ALL_BLOCKS_HELPERS + helpers.SNIPPET_STR_HELPERS)
    for helper_dets in all_helpers_dets:
        logging.debug(f"About to process '{helper_dets.helper_name}'")
        if warnings_only and not helper_dets.warning:
            continue
        if helper_dets.input_type == conf.BLOCKS_DETS:
            helper_input = blocks_dets
        elif helper_dets.input_type == conf.SNIPPET_STR:
            helper_input = snippet
        else:
            raise Exception(
                f"Unexpected input_type: '{helper_dets.input_type}'")
        repeat = (helper_dets.helper_name in repeat_set)
        message_dets = get_message_dets_from_input(helper_dets,
            helper_input=helper_input, code_str=snippet, first_line_no=None,
            repeat=repeat)
        if message_dets:
            repeat_set.add(helper_dets.helper_name)
            messages_dets.append(message_dets)
    return messages_dets

def get_separated_messages_dets(snippet, snippet_block_els, xml, *,
        warnings_only=False, repeat_set=None):
    """
    Break snippet up into syntactical parts and blocks of code. Apply helper
    functions and get message details. Split into overall messages and block-
    specific messages.

    :param str snippet: code snippet
    :param list snippet_block_els: list of block elements for snippet
    :param xml xml: snippet code as xml object
    :param bool warnings_only: if True, warnings only
    :param set repeat_set: we need to track if a help message is a repeat
     especially across multiple scripts being processed.
    :return: a tuple of two MessageDets lists
     (overall_snippet_messages_dets, block_level_messages_dets)
     or None if no messages
    :rtype: tuple (or None)
    """
    blocks_dets = get_blocks_dets(snippet, snippet_block_els)
    overall_snippet_messages_dets = get_overall_snippet_messages_dets(
        snippet, blocks_dets, warnings_only=warnings_only,
        repeat_set=repeat_set)
    block_level_messages_dets = get_block_level_messages_dets(
        blocks_dets, xml, warnings_only=warnings_only, repeat_set=repeat_set)
    for messages_dets in [
            overall_snippet_messages_dets, block_level_messages_dets]:
        if None in messages_dets:
            raise Exception("messages_dets in meant to be a list of "
                "MessageDets named tuples yet a None item was found")
    if not (overall_snippet_messages_dets or block_level_messages_dets):
        message = {
            conf.BRIEF: conf.NO_ADVICE_MESSAGE,
        }
        message = complete_message(message, source=conf.SYSTEM_MESSAGE)
        overall_snippet_messages_dets = [
            MessageDets(snippet, message,
                first_line_no=None, warning=False, source=conf.SYSTEM_MESSAGE)]
    return overall_snippet_messages_dets, block_level_messages_dets

def get_snippet_dets(snippet, *, warnings_only=False, repeat_set=None):
    """
    Get details for snippet of code.

    :return: snippet_messages_dets
     (overall_snippet_messages_dets, block_level_messages_dets),
     multi_block_snippet (bool)
    :rtype: tuple
    """
    tree = get_tree(snippet)
    xml = xml_from_tree(tree)
    if conf.RECORD_AST:
        ast_funcs.store_ast_output(xml)
    snippet_block_els = xml.xpath('body')[0].getchildren()  ## [0] because there is only one body under root
    multi_block_snippet = len(snippet_block_els) > 1
    snippet_messages_dets = get_separated_messages_dets(
        snippet, snippet_block_els, xml, warnings_only=warnings_only,
        repeat_set=repeat_set)
    return snippet_messages_dets, multi_block_snippet

def get_system_messages_dets(snippet, brief_message, *, warning=True):
    """
    Even though only one message is needed, supplying the details in the
    standard format the displayers expect means they can operate in their usual
    messages_dets consuming ways :-).
    """
    message = {
        conf.BRIEF: brief_message,
    }
    message = complete_message(message, source=conf.SYSTEM_MESSAGE)
    overall_messages_dets = [
        MessageDets(snippet, message,
            first_line_no=None, warning=warning, source=conf.SYSTEM_MESSAGE)]
    block_messages_dets = []
    messages_dets = (overall_messages_dets, block_messages_dets)
    return messages_dets

def get_error_messages_dets(e, snippet):
    """
    If unable to produce any messages, supply the problem in the form of
    standard messages_dets so the displayers can operate in their usual
    messages_dets consuming ways :-).
    """
    brief_message = layout(f"""\
        ### No advice sorry :-(

        Unable to provide advice - some sort of problem.

        Details: {e}
        """)
    return get_system_messages_dets(snippet, brief_message)

def get_community_message(snippet):
    brief_message = layout("""\

        ### Join in!

        Python has always had a great community. Learn more at
        <https://www.python.org/community/>. Better still - get involved :-)

        """)
    return get_system_messages_dets(snippet, brief_message)

def get_xkcd_warning(snippet):
    brief_message = layout("""\

        ### According to XKCD this code could be *very* dangerous

        See <https://xkcd.com/2261/>

        """)
    return get_system_messages_dets(snippet, brief_message)



import logging
from textwrap import dedent

from .cli_extras import md2cli
from .cli_extras.cli_colour import set_global_colours
from superhelp.gen_utils import (get_code_desc, get_intro,
    get_line_numbered_snippet, layout_comment as layout)

"""
Note - displays properly in the terminal but not necessarily in other output
e.g. Eclipse console.

Lots in common with md displayer but risks of DRYing probably outweigh benefits
at this stage.
"""

from .. import conf

TERMINAL_WIDTH = 80

MDV_CODE_BOUNDARY = "```"

def get_message(message_dets, detail_level):
    message = dedent(message_dets.message[detail_level])
    if detail_level == conf.EXTRA:
        message = dedent(message_dets.message[conf.MAIN]) + message
    message = dedent(message)
    message = (message
        .replace(f"    {conf.PYTHON_CODE_START}", MDV_CODE_BOUNDARY)
        .replace(f"\n    {conf.PYTHON_CODE_END}", MDV_CODE_BOUNDARY)
    )
    message = md2cli.main(md=message.replace('`', ''))  ## They create problems in formatting
    return message

def _need_snippet_displayed(overall_messages_dets, block_messages_dets, *,
        multi_block=False):
    """
    Don't need to see the code snippet displayed when it is already visible:
    * because there is only one block in snippet and there is a block message
      for it (which will display the block i.e. the entire snippet) UNLESS there
      is an overall message separating them
    Otherwise we need it displayed.
    """
    mono_block_snippet = not multi_block
    if mono_block_snippet and block_messages_dets and not overall_messages_dets:
        return False
    return True

def display(snippet, file_path, messages_dets, *,
        detail_level=conf.BRIEF,
        warnings_only=False, multi_block=False, multi_script=False,
        theme_name=None):
    """
    Show by code blocks.
    """
    set_global_colours(theme_name)
    md2cli.term_columns = TERMINAL_WIDTH
    if warnings_only:
        options_msg = conf.WARNINGS_ONLY_MSG
    else:
        options_msg = conf.ALL_HELP_SHOWING_MSG
    intro = get_intro(file_path, multi_block=multi_block)
    text = [
        md2cli.main(layout(f"""\
            # SuperHELP - Help for Humans!

            {intro}

            Currently showing {detail_level} content as requested.
            {options_msg}.

            {conf.MISSING_ADVICE_MESSAGE}

            ## Please give us a star &starf; on
            https://github.com/grantps/superhelp{conf.FORCE_SPLIT}
            so others can learn about us. Thanks!
            """
        )),
    ]
    overall_messages_dets, block_messages_dets = messages_dets
    display_snippet = _need_snippet_displayed(
        overall_messages_dets, block_messages_dets, multi_block=multi_block)
    if display_snippet:
        line_numbered_snippet = get_line_numbered_snippet(snippet)
        code_desc = get_code_desc(file_path)
        text.append(md2cli.main(dedent(
            f"## {code_desc}"
            f"\n{MDV_CODE_BOUNDARY}\n"
            + line_numbered_snippet
            + f"\n{MDV_CODE_BOUNDARY}")))
    for message_dets in overall_messages_dets:
        message = get_message(message_dets, detail_level)
        text.append(message)
    block_messages_dets.sort(key=lambda nt: (nt.first_line_no, nt.warning))
    prev_line_no = None
    for message_dets in block_messages_dets:
        ## display code for line number (once ;-))
        line_no = message_dets.first_line_no
        new_block = (line_no != prev_line_no)
        if new_block:
            block_has_warning_header = False
            text.append(md2cli.main(dedent(
                f'## Code block starting line {line_no:,}'
                f"\n{MDV_CODE_BOUNDARY}\n"
                + message_dets.code_str
                + f"\n{MDV_CODE_BOUNDARY}")))
            prev_line_no = line_no
        if message_dets.warning and not block_has_warning_header:
            text.append(md2cli.main(layout("""\
                ### Questions / Warnings

                There may be some issues with this code block you want to
                address.
                """)))
            block_has_warning_header = True
        ## process message
        message = get_message(message_dets, detail_level)
        text.append(message)
    content = '\n'.join(text)
    print(content)
    if multi_script:
        input("Press any key to continue ...")




from superhelp.helpers import all_blocks_help
from .. import conf
from superhelp.gen_utils import layout_comment as layout

def _includes_print(block_el):
    """
    print('Hi') is Call/func/Name id = 'print'
    p = print is Assign/value/Name id = 'print'
    """
    func_name_els = block_el.xpath('descendant-or-self::Call/func/Name')
    print_els = [func_name_el for func_name_el in func_name_els
        if func_name_el.get('id') == 'print']
    if print_els:
        return True
    assigned_name_els = block_el.xpath('descendant-or-self::Assign/value/Name')
    print_assigned_els = [
        assigned_name_el for assigned_name_el in assigned_name_els
        if assigned_name_el.get('id') == 'print']
    if print_assigned_els:
        return True
    return False

@all_blocks_help()
def print_overview(blocks_dets, *, repeat=False):
    """
    Show some of the surprise features of the humble print function.
    """
    has_print = False
    for block_dets in blocks_dets:
        if _includes_print(block_dets.element):
            has_print = True
            break
    if not has_print:
        return None

    title = layout("""\
    ### `print` function used
    """)
    if not repeat:
        brief_details = layout("""\
        `print` may seem like a very basic function but it has some extra
        parameters providing extra functionality. More information can be found
        in the official on-line documentation for `print`.
        """)
        main_details = (
            layout("""\

            `print` may seem like a very basic function but it has some extra
            features worth knowing about. Here is an example using the `end` and
            `flush` parameters. Dots should appear one by one in the same line.
            If `flush` were `False` then they would only appear at the end all
            at once. If `end` were the default new line character '\\n' they
            would have appeared on separate lines.
            """)
            +
            layout("""\
            from time import sleep
            print('.', end='', flush=True)
            sleep(0.25)
            """, is_code=True)
            +
            layout("""\

            More information can be found in `print`'s official on-line
            documentation.
            """)
        )
    else:
        brief_details = ''
        main_details = ''

    message = {
        conf.BRIEF: title + brief_details,
        conf.MAIN: title + main_details,
    }
    return message



from superhelp.helpers import get_dict_comprehension_msg, \
    get_general_comprehension_msg, get_set_comprehension_msg, filt_block_help
from ..ast_funcs import get_el_lines_dets
from .. import conf
from superhelp.gen_utils import layout_comment as layout

FOR_XPATH = 'descendant-or-self::For'

@filt_block_help(xpath=FOR_XPATH)
def comprehension_option(block_dets, *, repeat=False):
    """
    Provide overview of for loop to see if simple enough to be a possible
    candidate for a comprehension.

    Don't try to properly understand the for loop or make a built comprehension.
    It is enough to detect whether a loop is simple enough to consider making a
    comprehension or not. And to see whether appending, key setting, or adding
    is happening and suggesting the right comprehension accordingly.
    """
    for_els = block_dets.element.xpath(FOR_XPATH)
    any_short_enough = False
    for for_el in for_els:
        _first_line_no, _last_line_no, for_lines_n = get_el_lines_dets(
            for_el, ignore_trailing_lines=True)
        short_enough = for_lines_n < 3
        if short_enough:
            any_short_enough = True
            break
    if not any_short_enough:
        return None
    comp_type = None
    comp_comment = ''
    if 'append' in block_dets.block_code_str:
        comp_type = 'List Comprehension'
        comp_comment = get_dict_comprehension_msg()
    elif len(block_dets.element.cssselect('Subscript')):  ## Seems a reasonable indicator
        comp_type = 'Dictionary Comprehension'
        comp_comment = get_dict_comprehension_msg()
    elif 'set' in block_dets.block_code_str:
        comp_type = 'Set Comprehension'
        comp_comment = get_set_comprehension_msg()
    else:
        return None

    title = layout(f"""\
    ### Possible option of using a {comp_type}
    """)
    if not repeat:
        option = layout(f"""\

        Simple for loops can sometimes be replaced with comprehensions. In this
        case a simple reading of the code suggests a {comp_type} might be
        possible. Of course, only use a comprehension if it makes your code
        easier to understand.
        """)
    else:
        option = ''

    message = {
        conf.BRIEF: title + option,
        conf.MAIN: (title + option + get_general_comprehension_msg()
            + '\n\n' + comp_comment),
    }
    return message

def get_incremental_iteration_dets(for_el):
    """
    Look for the pattern:

    pets = ['cat', 'dog']
    for i in range(len(pets)):
        print(f"My {pets[i]}")

    For/target/Name id = i
        iter/Call/func/Name id = range
                  args/Call/func/Name id = len
                            args/Name id = pets

    ... Subscript/value/Name id = pets
                  slice/Index/value/Name = i
    """
    target_name_els = for_el.xpath('target/Name')
    if not target_name_els:
        return None
    index_name = target_name_els[0].get('id')  ## e.g. 'i'
    iterator_func_name_els = for_el.xpath('iter/Call/func/Name')
    if not iterator_func_name_els:
        return None
    range_iterator_func = iterator_func_name_els[0].get('id') == 'range'  ## range()
    if not range_iterator_func:
        return None
    iter_arg_func_name_els = for_el.xpath('iter/Call/args/Call/func/Name')
    if not iter_arg_func_name_els:
        return None
    len_inside = (iter_arg_func_name_els[0].get('id') == 'len')  ## range(len())
    if not len_inside:
        return None
    len_arg_name_els = for_el.xpath('iter/Call/args/Call/args/Name')
    if not len_arg_name_els:
        return None
    iterable_name = len_arg_name_els[0].get('id')  ## e.g. pets
    if not iterable_name:
        return None
    subscript_name_els = for_el.xpath(
        'descendant-or-self::Subscript/value/Name')
    if not subscript_name_els:
        return None
    subscript_name = subscript_name_els[0].get('id')  ## e.g. pets
    slicing_iterable = (iterable_name == subscript_name)
    if not slicing_iterable:
        return None
    slice_index_name_els = for_el.xpath(
        'descendant-or-self::Subscript/slice/Index/value/Name')
    if not slice_index_name_els:
        return None
    slice_index_name = slice_index_name_els[0].get('id')
    slicing_by_range_val = (slice_index_name == index_name)
    if not slicing_by_range_val:
        return None
    return index_name, iterable_name

@filt_block_help(xpath=FOR_XPATH)
def for_index_iteration(block_dets, *, repeat=False):
    """
    Look to see if an opportunity for simple iteration available as more
    pythonic alternative to incremental indexing.

    Look for the pattern:

    for i in range(len(foo)):
        foo[i] detected
    """
    for_els = block_dets.element.xpath(FOR_XPATH)
    any_incremental_iteration = False
    for for_el in for_els:
        try:
            index_name, iterable_name = get_incremental_iteration_dets(for_el)
        except TypeError:
            continue
        else:
            any_incremental_iteration = True
            break
    if not any_incremental_iteration:
        return None

    summary = layout(f"""\
    ### Possible option of using direct iteration

    It looks like your snippet iterates through `{iterable_name}` using indexes.
    In Python you can iterate directly which is much easier.
    """)
    if not repeat:
        examples = (
            layout("""\
            For example, instead of:
            """)
            +
            layout(f"""\
            for {index_name} in range(len({iterable_name})):
                print({iterable_name}[{index_name}])
            """, is_code=True)
            +
            layout("""\
            you can directly iterate as follows:
            """)
            +
            layout(f"""\
            for item in {iterable_name}:  ## item should be replaced with a more useful name
                print(item)

            """, is_code=True)
            +
            layout("""\
            which is considered more pythonic i.e. good.
            """)
        )
    else:
        examples = ''

    message = {
        conf.BRIEF: summary + examples,
    }
    return message

@filt_block_help(xpath=FOR_XPATH, warning=True)
def for_else(block_dets, *, repeat=False):
    """
    Look for the for-else construct and warn about its safe usage.
    """
    for_els = block_dets.element.xpath(FOR_XPATH)
    has_for_else = False
    for for_el in for_els:
        for_else_els = for_el.xpath('orelse')
        non_empty_for_else_els = [el for el in for_else_els if el.getchildren()]
        if non_empty_for_else_els:
            has_for_else = True
            break
    if not has_for_else:
        return None

    title = layout("""\
    ### Ambiguous `for-else` language feature used
    """)
    if not repeat:
        problem = (
            layout("""\

            `for`-`else` is a useful language feature but it is so widely
            misunderstood it is dangerous to use without special comment. For
            example, in the snippet below do we pass through `else` because we
            finished the looping or because we didn't?
            """)
            +
            layout("""\
            for i in range(size):
                if i > 100:
                    break
            else:
                pass
            """, is_code=True)
        )
        solution = (
            layout("""\

            The correct answer, which as many as a third of people will guess
            correctly ;-), is that we only follow the `else` path if we finished
            the for-looping without breaking. The thinking is that we loop
            through something to find something and exit the loop. `else`
            applies when that search fails. But it is very easy to get the logic
            confused. And given that programs are made to be read by humans we
            obviously need to do something. We can either avoid the feature
            altogether or we can add a brief comment each time (my preference).
            For example:
            """)
            +
            layout("""\
            for i in range(size):
                if i > 100:
                    break
            else:  # didn't break for-loop
                pass
            """, is_code=True)
        )
        extra_msg = layout("""\
        #### Interesting discussion on `for`-`else`

        <https://nedbatchelder.com/blog/201110/forelse.html>
        One comment noted that the construct violates the principle of Least
        Astonishment.
        """)
    else:
        problem = ''
        solution = ''
        extra_msg = ''

    message = {
        conf.BRIEF: title + problem,
        conf.MAIN: title + problem + solution,
        conf.EXTRA: extra_msg,
    }
    return message

@filt_block_help(xpath=FOR_XPATH)
def nested_fors(block_dets, *, repeat=False):
    """
    Look to see if an opportunity for using itertools.product instead of nested
    iteration.
    """
    for_els = block_dets.element.xpath(FOR_XPATH)
    nested_iteration = False
    for for_el in for_els:
        nested_for_els = for_el.xpath('descendant::For')
        if nested_for_els:
            nested_iteration = True
            break
    if not nested_iteration:
        return None

    summary = layout("""\
    ### Possible option of simplifying nested iteration

    Consider replacing nested iteration with `itertools.product`.
    """)
    if not repeat:
        demo = (
            layout("""\
            For example, you could replace:
            """)
            +
            layout("""\
            for person in persons:
                for pet in pets:
                    for year in years:
                        print(f"{person} might like a {pet} in {year}")
            """, is_code=True)
            +
            layout("""\
            with a version using `itertools.product`:
            """)
            +
            layout("""\
            from itertools import product
            for person, pet, year in product(persons, pets, years):
                print(f"{person} might like a {pet} in {year}")
            """, is_code=True)
        )
        pros = layout("""\

        Whether this is a good idea or not depends on your specific code but
        using `product` has the advantage of reducing indentation. It also
        semantically expresses the intention of the code - namely to look at
        every option in the cartesian product of the different collections of
        items.
        """)
    else:
        demo = ''
        pros = ''

    message = {
        conf.BRIEF: summary + demo,
        conf.MAIN: summary + demo + pros,
    }
    return message



"""
Covers functions and methods.
"""
from superhelp.helpers import filt_block_help
from ..ast_funcs import get_danger_status, get_docstring_from_value, \
    get_el_lines_dets
from .. import conf
from superhelp import gen_utils
from superhelp.gen_utils import get_nice_pairs, layout_comment as layout

FUNC_DEFN_XPATH = 'descendant-or-self::FunctionDef'

def get_mutable_status(default_el):
    """
    KISS - Only look for List, Dict, and Set. Python 3.6 as well as 3.8
    compatible.
    """
    if default_el.tag in ['List', 'Dict', 'Set']:
        mutable_status = default_el.tag
    else:
        mutable_status = None
    return mutable_status

def get_is_method(func_el):
    class_els = func_el.xpath('ancestor::ClassDef')
    if not class_els:
        return False
    class_el = class_els[-1]
    ## check is a direct member of a class and not just a function defined somewhere internally
    method_els = class_el.xpath('body/FunctionDef')
    return func_el in method_els

def get_func_type_lbl(func_el):
    func_type_lbl = (
        conf.METHOD_LBL if get_is_method(func_el) else conf.FUNCTION_LBL)
    return func_type_lbl

def get_overall_func_type_lbl(func_els):
    if not func_els:
        return None
    includes_plain_function = False
    for func_el in func_els:
        func_type_lbl = get_func_type_lbl(func_el)
        if func_type_lbl == conf.FUNCTION_LBL:
            includes_plain_function = True
    overall_func_type_lbl = (
        conf.FUNCTION_LBL if includes_plain_function else conf.METHOD_LBL)
    return overall_func_type_lbl

def count_args(func_el):
    """
    Have to handle positional-only, keyword-only, and standard arguments. Not
    counting any unpacked args (vararg and kwarg).
    """
    posonlyargs = func_el.xpath('args/arguments/posonlyargs/arg')  ## python 3.8+
    args = func_el.xpath('args/arguments/args/arg')
    kwonlyargs = func_el.xpath('args/arguments/kwonlyargs/arg')
    all_args_n = len(posonlyargs + args + kwonlyargs)
    return all_args_n

def _get_arg_comment(func_el, *, repeat=False):
    """
    Must cope with positional-only arguments, positional arguments, keyword
    arguments, keyword-only arguments, packed positional arguments, and packed
    keyword arguments. Trivial really ;-).

    Comment should end without a full stop because calling code adds that to
    make the sentence structure more explicit.
    """
    vararg = func_el.xpath('args/arguments/vararg/arg')
    kwarg = func_el.xpath('args/arguments/kwarg/arg')
    has_packing = (vararg or kwarg)
    if has_packing:
        arg_comment = 'receives a variable number of arguments'
        if repeat:
            arg_comment += '.'
        else:
            if vararg:
                vararg_name = vararg[0].get('arg')
                arg_comment += (". All positional arguments received are packed"
                    f" together into a list called {vararg_name} using "
                    f"the &ast;{vararg_name} syntax. If there is no better name"
                    " in a particular case the Python convention is to call "
                    "that list 'args'")
            if kwarg:
                kwarg_name = kwarg[0].get('arg')
                arg_comment += (". All keyword arguments received are packed "
                    f"together into a dictionary called {kwarg_name} "
                    f"using the &ast;&ast;{kwarg_name} syntax. If there is no "
                    "better name in a particular case the Python convention is "
                    "to call that dictionary 'kwargs'")
    else:
        all_args_n = count_args(func_el)
        if all_args_n:
            nice_n_args = gen_utils.int2nice(all_args_n)
            arg_comment = (f"receives {nice_n_args} argument")
            if all_args_n > 1:
                arg_comment += 's'
        else:
            arg_comment = "doesn't take any arguments"
    return arg_comment

def _get_return_comment(func_type_lbl, return_elements, *,
        repeat=False):
    """
    Comment should end without a full stop because calling code adds that to
    make the sentence structure more explicit.
    """
    implicit_return_els = [return_element for return_element in return_elements
        if not return_element.getchildren()]
    implicit_returns_n = len(implicit_return_els)
    explicit_return_els = [return_element for return_element in return_elements
        if return_element.xpath('value')]
    none_returns_n = 0
    val_returns_n = 0
    for el in explicit_return_els:
        val_children = el.getchildren()
        for val_child in val_children:
            if (val_child.tag == 'NameConstant'
                    and val_child.get('value') is None):
                none_returns_n += 1
            else:
                val_returns_n += 1
    keyword_returns_n = none_returns_n + val_returns_n + implicit_returns_n
    if not keyword_returns_n:
        returns_comment = (
            f"The {func_type_lbl} does not explicitly return anything")
        if not repeat:
            returns_comment += (
                ". In which case, in Python, it implicitly returns `None`")
    else:
        returns_comment = (
            f"The {func_type_lbl} exits via an explicit `return` statement "
            f"{gen_utils.int2nice(keyword_returns_n)} time")
        if not repeat:
            if keyword_returns_n > 1:
                returns_comment += (
                f"s. Some people prefer {func_type_lbl}s to have one, and only "
                "one exit point for clarity. Others disagree e.g. using early "
                f"returns to short-circuit {func_type_lbl}s if initial "
                "validation of some sort makes it possible to avoid the bulk of"
                f" the {func_type_lbl}. Whatever approach you take make sure "
                f"your {func_type_lbl} is easy to reason about in terms of what"
                " it returns and where it exits")
    return returns_comment

def _get_exit_comment(func_el, func_type_lbl, *, repeat=False):
    """
    Look for 'return' and 'yield'.
    """
    return_elements = func_el.xpath('descendant-or-self::Return')
    yield_elements = func_el.xpath('descendant-or-self::Yield')
    if yield_elements:
        if return_elements:
            exit_comment = ("It has both `return` and `yield`. "
                "That probably doesn't make any sense.")
        else:
            exit_comment = "It is a generator function."
    else:
        exit_comment = _get_return_comment(func_type_lbl,
            return_elements, repeat=repeat)
    return exit_comment

@filt_block_help(xpath=FUNC_DEFN_XPATH)
def func_overview(block_dets, *, repeat=False):
    """
    Advise on function (or method) definition statements.
    e.g. def greeting(): ...
    """
    func_els = block_dets.element.xpath(FUNC_DEFN_XPATH)
    overall_func_type_lbl = get_overall_func_type_lbl(func_els)

    title = layout(f"""\

        ### {overall_func_type_lbl.title()} Details
        """)
    detail_bits = []
    for func_el in func_els:
        func_type_lbl = get_func_type_lbl(func_el)
        name = func_el.get('name')
        arg_comment = _get_arg_comment(
            func_el, repeat=repeat)
        exit_comment = _get_exit_comment(
            func_el, func_type_lbl, repeat=repeat)
        detail_bits.append(layout(f"""\

            The {func_type_lbl} named `{name}` {arg_comment}. {exit_comment}.
            """))
    details = ''.join(detail_bits)
    if not repeat:
        args_vs_params = layout(f"""\

        There is often confusion about the difference between arguments and
        parameters. {overall_func_type_lbl.title()}s define parameters but
        receive arguments. You can think of parameters as being like car parks
        and arguments as the cars that fill them. You supply arguments to a
        {overall_func_type_lbl} depending on its parameters.
        """)
        if func_type_lbl == conf.METHOD_LBL:
            methods = layout("""\

            Methods are functions that sit directly inside a class definition.
            Unless they are defined as static methods e.g. using the
            `@staticmethod` decorator, they take the instance of the class as
            the first parameter - almost always named `self`. But they are
            basically functions.
                """)
        else:
            methods = ''
    else:
        args_vs_params = ''
        methods = ''

    message = {
        conf.BRIEF: title + details,
        conf.MAIN: title + details + methods,
        conf.EXTRA: args_vs_params,
    }
    return message

@filt_block_help(xpath=FUNC_DEFN_XPATH, warning=True)
def func_len_check(block_dets, *, repeat=False):
    """
    Warn about functions that might be too long.
    """
    func_els = block_dets.element.xpath(FUNC_DEFN_XPATH)
    overall_func_type_lbl = get_overall_func_type_lbl(func_els)
    long_func_dets = []
    for func_el in func_els:
        func_type_lbl = get_func_type_lbl(func_el)
        name = func_el.get('name')
        first_line_no, last_line_no, _func_lines_n = get_el_lines_dets(
            func_el, ignore_trailing_lines=True)
        block_lines = block_dets.block_code_str.split('\n')
        func_lines = block_lines[first_line_no - 1: last_line_no]
        func_non_empty_lines = [line for line in func_lines if line]
        func_lines_n = len(func_non_empty_lines)
        if func_lines_n <= conf.MAX_BRIEF_FUNC_LOC:
            continue
        else:
            long_func_dets.append((name, func_lines_n))
    if not long_func_dets:
        return None

    title = layout(f"""\
    ### {overall_func_type_lbl.title()} possibly too long
    """)
    summary_bits = []
    for name, func_lines_n in long_func_dets:
        summary_bits.append(layout(f"""\

        `{name}` has {gen_utils.int2nice(func_lines_n)} lines of code (including
        comments but with empty lines ignored).
        """))
    summary = ''.join(summary_bits)
    if not repeat:
        sometimes_ok = (
            f" Sometimes it is OK for a {func_type_lbl} to be "
            "that long but you should consider refactoring the code "
            "into smaller units.")
    else:
        sometimes_ok = ''

    message = {
        conf.BRIEF: title + summary + sometimes_ok,
    }
    return message

def get_n_args(func_el):
    arg_els = func_el.xpath('args/arguments/args/arg')
    posonlyarg_els = func_el.xpath('args/arguments/posonlyargs/arg')
    kwonlyarg_els = func_el.xpath('args/arguments/kwonlyargs/arg')
    n_args = len(arg_els + posonlyarg_els + kwonlyarg_els)
    return n_args

@filt_block_help(xpath=FUNC_DEFN_XPATH, warning=True)
def func_excess_parameters(block_dets, *, repeat=False):
    """
    Warn about functions that might have too many parameters.
    """
    func_els = block_dets.element.xpath(FUNC_DEFN_XPATH)
    overall_func_type_lbl = get_overall_func_type_lbl(func_els)
    excess_param_dets = []
    for func_el in func_els:
        func_type_lbl = get_func_type_lbl(func_el)
        name = func_el.get('name')
        n_args = get_n_args(func_el)
        high_args = n_args > conf.MAX_BRIEF_FUNC_ARGS
        if high_args:
            excess_param_dets.append((name, n_args, func_type_lbl))
    if not excess_param_dets:
        return None

    title = layout(f"""\
    ### Possibly too many {overall_func_type_lbl} parameters
    """)
    summary_bits = []
    for i, (name, n_args, func_type_lbl) in enumerate(excess_param_dets):
        first = (i == 0)
        summary_bits.append(layout(f"""\

        `{name}` has {n_args:,} parameters.
        """))
        if first and not repeat:
            summary_bits.append(layout(f"""\

            Sometimes it is OK for a {func_type_lbl} to have that many but you
            should consider refactoring the code or collecting related
            parameters into single parameters e.g. instead of receiving image
            size arguments separately perhaps you could receive a dictionary of
            image size argument details.
            """))
    summary = ''.join(summary_bits)

    message = {
        conf.BRIEF: title + summary,
    }
    return message

def get_arg_default_issues(func_el, *, get_issue_status_func, include_kw=True):
    """
    Look at this function's arguments. Any issues?
    """
    posonly_arg_els = func_el.xpath('args/arguments/posonlyargs/arg')
    arg_els = func_el.xpath('args/arguments/args/arg')
    all_arg_els = posonly_arg_els + arg_els  ## order matters
    if include_kw:
        kwonly_arg_els = func_el.xpath('args/arguments/kwonlyargs/arg')
        all_arg_els +=  kwonly_arg_els
    arg_names = [arg_el.get('arg') for arg_el in all_arg_els]
    arg_default_els = func_el.xpath('args/arguments/defaults')
    default_els = arg_default_els  ## order matters
    if include_kw:
        kw_default_els = func_el.xpath('args/arguments/kw_defaults')
        default_els += kw_default_els
    issue_statuses = []
    for default_el in default_els:
        for child_el in default_el.getchildren():
            issue_status = get_issue_status_func(child_el)
            issue_statuses.append(issue_status)
    ## reversed because defaults are filled in rightwards e.g. a, b=1, c=2
    ## args = a,b,c and defaults=1,2 -> reversed c,b,a and 2,1 -> c: 2, b: 1
    arg_names_reversed = reversed(arg_names)
    issue_statuses_reversed = reversed(issue_statuses)
    args_and_issue_statuses = reversed(list(
        zip(arg_names_reversed, issue_statuses_reversed)))  ## back to left-to-right order
    args_with_issues = [(arg, issue_status)
        for arg, issue_status in args_and_issue_statuses
        if issue_status]
    return args_with_issues

def _get_mutable_default_args(func_el):
    """
    Interested in all args, not just non-keyword ones as when looking for danger
    arguments. We are also interested in all defaults as well.
    """
    return get_arg_default_issues(
        func_el, get_issue_status_func=get_mutable_status, include_kw=True)

@filt_block_help(xpath=FUNC_DEFN_XPATH, warning=True)
def mutable_default(block_dets, *, repeat=False):
    """
    Look for use of mutable defaults and warn against use except in rare cases.
    """
    func_els = block_dets.element.xpath(FUNC_DEFN_XPATH)
    overall_func_type_lbl = get_overall_func_type_lbl(func_els)
    mutable_defaults_dets = []
    for func_el in func_els:
        func_type_lbl = get_func_type_lbl(func_el)
        name = func_el.get('name')
        mutable_default_args = _get_mutable_default_args(func_el)
        if mutable_default_args:
            mutable_defaults_dets.append(
                (name, func_type_lbl, mutable_default_args))
    if not mutable_defaults_dets:
        return None

    title = layout(f"""\
    ### {overall_func_type_lbl.title()} has mutable default arguments
    """)
    brief_summary_bits = []
    main_summary_bits = []
    for i, mutable_default_dets in enumerate(mutable_defaults_dets):
        name, func_type_lbl, mutable_default_args = mutable_default_dets
        first = (i == 0)
        nice_pairs = get_nice_pairs(
            mutable_default_args, pair_glue=' defaults to a ')
        issue = layout(f"""\

        `{name}` has the following parameters with mutable defaults:
        {nice_pairs}.
        """)
        brief_summary_bits.append(issue)
        main_summary_bits.append(issue)
        if first and not repeat:  ## explaining once is enough ;-)
            main_summary_bits.append(layout(f"""\

            Mutable default arguments are a well-known "gotcha" in Python. They
            do not behave as most people would expect them to. It is better to
            have `None` as the default, and inside the function set the name to
            the desired mutable if it is `None`.
            """)
            +
            layout("""\
            ## BAD (mutable default)
            def communicate(person, msg, people=[]):
                people.append(person)
                for person in people:
                    print(f"Emailed {person} with message: '{msg}'")

            ## GOOD (no mutable defaults)
            def communicate(person, msg, people=None):
                if people is None:
                    people = []
                people.append(person)
                for person in people:
                    print(f"Emailed {person} with message: '{msg}'")
            """, is_code=True)
            )
    brief_summary = ''.join(brief_summary_bits)
    main_summary = ''.join(main_summary_bits)
    if not repeat:
        defaults_explained = (
            layout(f"""\

            {overall_func_type_lbl.title()}s are stored in memory when they are
            defined, and that includes any default arguments that belong to
            them. So every time the {overall_func_type_lbl} is called it is from
            the same piece of memory and the default argument is the same.
            Calling a {overall_func_type_lbl} multiple times will mean starting
            from the mutable default as it was left by the previous call - which
            makes sense because it is one and the same thing.

            Let's use our BAD version of the `communicate` function to see what
            happens:
            """)
            +
            layout("""\
            ## BAD (people list will persist across multiple calls)
            def communicate(person, msg, people=[]):
                people.append(person)
                for person in people:
                    print(f"Emailed {person} with message: '{msg}'")

            communicate('Jack', 'Apples are on special')
            ## people is empty, so only Jack is emailed
            # >>> Emailed Jack with message: 'Apples are on special'

            communicate('Jill', 'Tennis this Saturday')
            ## people started with Jack in it so both Jack AND Jill get emailed!
            # >>> Emailed Jack with message: 'Tennis this Saturday'  # <---oops!
            # >>> Emailed Jill with message: 'Tennis this Saturday'
            """, is_code=True)
        )
    else:
        defaults_explained = ''

    message = {
        conf.BRIEF: title + brief_summary,
        conf.MAIN: title + main_summary + defaults_explained,
    }
    return message

def _get_positional_danger_args(func_el):
    """
    Interested in non-keyword args only. So posonly_args and args respectively.
    As for defaults, defaults only (ignoring kw_defaults - the only other option
    - there is no separate posonly_defaults)
    """
    return get_arg_default_issues(
        func_el, get_issue_status_func=get_danger_status, include_kw=False)

@filt_block_help(xpath=FUNC_DEFN_XPATH, warning=True)
def positional_boolean(block_dets, *, repeat=False):
    """
    Look for any obvious candidates for forced keyword use e.g. where a
    parameter is a boolean or a number.

    Defaults apply from the rightmost backwards (within their group - either
    defaults or kw_defaults (related to kwonlyargs)).
    """
    func_els = block_dets.element.xpath(FUNC_DEFN_XPATH)
    overall_func_type_lbl = get_overall_func_type_lbl(func_els)
    positional_dets = []
    for func_el in func_els:
        func_type_lbl = get_func_type_lbl(func_el)
        name = func_el.get('name')
        danger_args = _get_positional_danger_args(func_el)
        if danger_args:
            positional_dets.append((name, func_type_lbl, danger_args))
    if not positional_dets:
        return None

    title = layout(f"""\
    ### {overall_func_type_lbl.title()} expects risky positional arguments
    """)
    summary_bits = []
    for i, (name, func_type_lbl, danger_args) in enumerate(positional_dets):
        first = (i == 0)
        summary_bits.append(layout(f"""\

        A partial analysis of `{name}` found the following risky non-keyword
        (positional) parameters: {get_nice_pairs(danger_args, left_quoter='')}.
        """))
        if first and not repeat:  ## explaining once is enough ;-)
            summary_bits.append(layout(f"""\

            {func_type_lbl.title}s which expect numbers or booleans
            (`True`/`False`) without requiring keywords are risky. They are
            risky when if the {func_type_lbl} is changed later to have different
            parameters. For example, `greeting(formal=True)` is more
            intelligible than `greeting(True)`. And intelligible code is safer
            to alter / maintain over time than mysterious code.
            """))
    summary = ''.join(summary_bits)
    if not repeat:
        asterisk_demo = (
            layout("""\

            Using an asterisk as a pseudo-parameter forces all parameters to the
            right to be keywords e.g.
            """)
            +
            layout("""\
            def greeting(name, *, formal=False):
                ...
            """, is_code=True)
            +
            layout("""\

            In this example you couldn't now call the function `greeting('Jo',
            True)` - it would need to be `greeting('Jo', formal=True)`
            """)
        )
        asterisk_explained = layout(f"""\

        Putting an asterisk in the parameters has the effect of forcing all
        parameters to the right to be keyword parameters because the asterisk
        mops up any remaining positional arguments supplied (if any) when the
        {func_type_lbl} is called. There can't be any other positional
        arguments, because they have all been handled already, so only keyword
        parameters are allowed thereafter.
        """)
    else:
        asterisk_demo = ''
        asterisk_explained = ''

    message = {
        conf.BRIEF: title + summary + asterisk_demo,
        conf.EXTRA: asterisk_explained,
    }
    return message

def get_func_name_docstring(func_el):
    func_body_el = func_el.xpath('body')[0]
    func_name = func_el.get('name')
    ## first item in body MUST be Expr and the first value must be a Str
    body_els = func_body_el.getchildren()
    if not body_els:
        docstring = None
    else:
        first_body_el = body_els[0]
        if first_body_el.tag != 'Expr':
            docstring = None
        else:
            expr_els = first_body_el.getchildren()
            if not expr_els:
                docstring = None
            else:
                first_expr_el = expr_els[0]
                if first_expr_el.tag != 'value':
                    docstring = None
                else:
                    value_els = first_expr_el.getchildren()
                    if not value_els:
                        docstring = None
                    else:
                        first_value_el = value_els[0]
                        docstring = get_docstring_from_value(first_value_el)
    return func_name, docstring

def get_funcs_dets_and_docstring(func_els):
    funcs_dets_and_docstring = []
    for func_el in func_els:
        func_name, docstring = get_func_name_docstring(func_el)
        func_dets_and_docstring = (func_el, func_name, docstring)
        funcs_dets_and_docstring.append(func_dets_and_docstring)
    return funcs_dets_and_docstring

@filt_block_help(xpath=FUNC_DEFN_XPATH, warning=True)
def docstring_issues(block_dets, *, repeat=False):
    """
    Check over function doc strings. Missing doc string, not enough lines to
    cover params, return etc.
    """
    WRAPPING_NEWLINE_N = 2
    MISSING_DOCSTRING = 'missing_docstring'
    DOCSTRING_TOO_SHORT = 'docstring_too_short'
    func_els = block_dets.element.xpath(FUNC_DEFN_XPATH)
    funcs_dets_and_docstring = get_funcs_dets_and_docstring(func_els)
    docstring_issues = []
    for func_el, func_name, docstring in funcs_dets_and_docstring:
        func_type_lbl = get_func_type_lbl(func_el)
        if docstring is None:
            docstring_issues.append(
                (func_name, func_type_lbl, MISSING_DOCSTRING))
        else:
            n_args = get_n_args(func_el)
            n_doc_lines = len(docstring.split('\n')) - WRAPPING_NEWLINE_N
            too_short = n_doc_lines < (conf.MIN_BRIEF_DOCSTRING + n_args)
            param_str = ' given the number of parameters' if n_args > 1 else ''
            if too_short:
                docstring_issues.append(
                (func_name, func_type_lbl, DOCSTRING_TOO_SHORT))
    if not docstring_issues:
        return None

    title = layout("""\
    ### Function / Method missing doc string
    """)
    example_docstring = layout(f'''\
    def greet(name, greet_word='Hi'):
        """
        Get a greeting for the supplied person.

        :param str name: person being greeted
        :param str greet_word: the word to start the greeting
        :return: a greeting message to the person
        :rtype: str
        """
        greeting = f"{{greet_word}} {{name}} - how are you?"
        return greeting
    ''', is_code=True)
    summary_bits = []
    for i, (func_name, func_type_lbl, problem) in enumerate(docstring_issues):
        first = (i == 0)
        if problem == MISSING_DOCSTRING:
            if first and not repeat:  ## only want to say it once ;-)
                summary_bits.append((
                    layout(f"""\

                    #### {func_type_lbl.title()} missing doc string

                    `{func_name}` lacks a doc string - you should probably add
                    one.

                    Note - # comments at the top of the {func_type_lbl} do not
                    work as doc strings. Python completely ignores them. If you
                    add a proper doc string, however, it can be accessed by
                    running `help({func_name})` or `{func_name}.`\_\_doc\_\_.
                    Which is useful when using this {func_type_lbl} in bigger
                    projects e.g. in an IDE (Integrated Development
                    Environment).

                    Here is an example doc string for a simple function using
                    one of several valid formats:
                    """)
                    +
                    example_docstring
                ))
            else:
                summary_bits.append(layout(f"""\
                #### `{func_name}` lacks a doc string

                You should probably add a doc string to `{func_name}`
                """))
        elif problem == DOCSTRING_TOO_SHORT:
            if first and not repeat:
                summary_bits.append((layout(f"""\
                #### Function doc string too brief?

                The doc string for `{func_name}` seems a little
                short{param_str}. You might want to rework it. Here is an
                example using one of several valid formats:
                """)
                +
                example_docstring
                ))
            else:
                summary_bits.append(layout(f"""\
                #### Function doc string too brief?

                The doc string for `{func_name}` seems a little short.
                """))
    summary = ''.join(summary_bits)

    message = {
        conf.BRIEF: title + summary,
    }
    return message



from superhelp.helpers import any_block_help, filt_block_help
from .. import conf, name_utils
from superhelp.gen_utils import get_nice_str_list, layout_comment as layout

def _get_sorting_or_reversing_comment(block_dets):
    """
    Get a comment on any sorting or reversing identified.

    :return: string describing type of reversing/sorting or None
    :rtype: str
    """
    element = block_dets.element
    comment = None
    func_attr_els = element.xpath('descendant-or-self::Call/func/Attribute')
    sort_els = [func_attr_el for func_attr_el in func_attr_els
        if func_attr_el.get('attr') == 'sort']
    func_name_els = element.xpath('descendant-or-self::Call/func/Name')
    sorted_els = [func_name_el for func_name_el in func_name_els
        if func_name_el.get('id') == 'sorted']
    reversed_els = [func_name_el for func_name_el in func_name_els
        if func_name_el.get('id') == 'reversed']
    if sort_els:
        comment = "has list sorting (`.sort()`)"
    if comment and (sorted_els or reversed_els):
        comment = ' and ' + comment
    if sorted_els and reversed_els:
        comment = "uses both the `sorted` and `reversed` functions"
    elif sorted_els:
        comment = "uses the `sorted` function"
    elif reversed_els:
        comment = "uses the `reversed` function"
    return comment

@any_block_help()
def sorting_reversing_overview(block_dets, *, repeat=False):
    """
    Provide an overview of sorting and/or reversing. Advise on common
    confusions.
    """
    sorting_or_reversing_comment = _get_sorting_or_reversing_comment(block_dets)
    if not sorting_or_reversing_comment:
        return None

    title = layout("""\
    ### Sorting / reversing
    """)
    if not repeat:
        summary = layout(f"""\

        This block of code {sorting_or_reversing_comment}.

        Sorting and, to a lesser extent, reversing are very common needs in
        programming. Two key points:

        1) reversing is not the same as sorting with `reverse=True`

        2) the list sort method e.g. my_list.`sort()` returns `None`, not the
        sorted list
        """)
        details = (
            layout(f"""\

            Sorting and, to a lesser extent, reversing are very common needs in
            programming. Two key points:

            1) reversing is not the same as sorting with `reverse=True`

            To illustrate:
            """)
            +
            layout("""\
            word = 'cat'
            word_reversed = reversed(word)
            ## >>> word_reversed
            ## >>> 'tac'
            word_reverse_sorted = sorted(word, reverse=True)
            ## >>> word_reverse_sorted
            ## >>> 'tca'
            ## >>> word_reversed == word_reverse_sorted
            ## >>> False
            """, is_code=True)
            +
            layout("""\

            Using the reversed function does not apply any sorting to the
            sequence being reversed - it merely flips the (possibly) unordered
            sequence the other way.

            2) the list sort method e.g. `my_list.sort()` returns `None`, not
            the sorted list

            `sorted(my_list)` returns a sorted list but `my_list.sort()` is an
            in-place process. It mutates something rather than returning a
            separate thing.

            To illustrate:

            i) `sorted()` returning a result and leaving its input unchanged

            """)
            +
            layout("""\
            fruit = ['banana', 'apple', 'cranberry']
            fruit_sorted = sorted(fruit)
            ## >>> fruit_sorted
            ## >>> ['apple', 'banana', 'cranberry']
            ## fruit itself has not been changed
            ## >>> fruit
            ## >>> ['banana', 'apple', 'cranberry']
            """, is_code=True)
            +
            layout("""\
            ii) `.sort()` returning `None` and changing its input in-place
            """)
            +
            layout("""\
            result_of_fruit_sort = fruit.sort()
            ## >>> result_of_fruit_sort
            ## >>> None
            ## fruit has been changed by the in-place sort method
            ## >>> fruit
            ## >>> ['apple', 'banana', 'cranberry']
            """, is_code=True)
            )
    else:
        summary = layout(f"""\
        This block of code {sorting_or_reversing_comment}.
        """)
        details = summary

    message = {
        conf.BRIEF: title + summary,
        conf.MAIN: title + details,
    }
    return message

ASSIGN_FUNC_ATTRIBUTE_XPATH = 'descendant-or-self::Assign/value/Call/func/Attribute'

@filt_block_help(xpath=ASSIGN_FUNC_ATTRIBUTE_XPATH, warning=True)
def list_sort_as_value(block_dets, *, repeat=False):
    """
    Warn about assigning a name to the result using .sort() on a list.
    """
    func_attr_els = block_dets.element.xpath(ASSIGN_FUNC_ATTRIBUTE_XPATH)
    names_assigned_to_sort = []
    for func_attr_el in func_attr_els:
        is_sort = (func_attr_el.get('attr') == 'sort')
        if is_sort:
            try:
                name_dets = name_utils.get_assigned_name(func_attr_el)
            except Exception:
                continue
            names_assigned_to_sort.append(name_dets.name_str)
    if not names_assigned_to_sort:
        return None

    title = layout("""\
    ### Assignment of `None` result from in-place `.sort()` on list
    """)
    if not repeat:
        multiple = len(names_assigned_to_sort) > 1
        if multiple:
            nice_str_list = get_nice_str_list(
                names_assigned_to_sort, quoter='`')
            details = layout(f"""\

            {nice_str_list} are assigned to the results of in-place sort
            operations. This is almost certainly a mistake as the intention is
            probably not to set them each to `None` (the return value of the
            `.sort()` method).
            """)
        else:
            name = names_assigned_to_sort[0]
            details = layout(f"""\

            `{name}` is assigned to the result of an in-place sort operation.
            This is almost certainly a mistake as the intention is probably not
            to set `{name}` to `None` (the return value of the `.sort()`
            method).
            """)
    else:
        details = ''

    message = {
        conf.BRIEF: title + details,
    }
    return message



from superhelp.helpers import all_blocks_help
from .. import conf
from superhelp.gen_utils import get_nice_str_list, int2nice, layout_comment as layout

UNSPECIFIC_EXCEPTION = 'Exception'

def get_exception_blocks(blocks_dets):
    """
    There can be multiple try-except statements in a snippet so we have to
    handle each of them.
    """
    exception_blocks = []
    for block_dets in blocks_dets:
        block_exception_types = []
        except_handler_els = block_dets.element.xpath(
            'descendant-or-self::handlers/ExceptHandler')
        for except_handler_el in except_handler_els:
            exception_type_els = except_handler_el.xpath(
                'type/Name | type/Tuple/elts/Name')
            exception_handler_types = [
                exception_type_el.get('id')
                for exception_type_el in exception_type_els]
            block_exception_types.extend(exception_handler_types)
        if block_exception_types:
            exception_blocks.append(block_exception_types)
    return exception_blocks

@all_blocks_help()
def exception_overview(blocks_dets):
    """
    Provide overview of exception handling.
    """
    exception_blocks = get_exception_blocks(blocks_dets)
    if not exception_blocks:
        return None

    title = layout("""\
    ### Exception handling
    """)
    block_comment_bits = []
    for n, exception_block in enumerate(exception_blocks, 1):
        counter = '' if len(exception_blocks) == 1 else f" {int2nice(n)}"
        handlers = get_nice_str_list(exception_block, quoter='`')
        block_comment_bits.append(layout(f"""\
        #### `try`-`except` block{counter}

        The following exception handlers were detected: {handlers}
        """))
    block_comments = ''.join(block_comment_bits)

    message = {
        conf.BRIEF: title + block_comments,
    }
    return message

@all_blocks_help(warning=True)
def unspecific_exception(blocks_dets):
    """
    Look for unspecific exceptions.
    """
    exception_blocks = get_exception_blocks(blocks_dets)
    if not exception_blocks:
        return None
    unspecific_block_ns = []
    for n, exception_block in enumerate(exception_blocks, 1):
        only_unspecific = (len(exception_block) == 1
            and exception_block[0] == UNSPECIFIC_EXCEPTION)
        if not only_unspecific:
            continue
        unspecific_block_ns.append(n)
    if not unspecific_block_ns:
        return None

    title = layout("""\
    #### Un-specific `Exception` only in `try`-`except` block(s)
    """)

    n_unspecific = len(unspecific_block_ns)
    if n_unspecific == 1:
        block_n_specific_text = ' has'
    else:
        unspecific_nice_block_ns = [
            int2nice(unspecific_block_n)
            for unspecific_block_n in unspecific_block_ns]
        blocks_ns = get_nice_str_list(unspecific_nice_block_ns, quoter='')
        block_n_specific_text = f"s {blocks_ns} have"
    unspecific_warning = layout(f"""\

    `try`-`except` block{block_n_specific_text} an un-specific Exception only.

    Using the un-specific exception type `Exception` is often completely
    appropriate. But if you are looking for specific exceptions you should
    handle those separately.
    """)
    unspecific_demo = (
        layout("""\
        For example:
        """)
        +
        layout("""\

        try:
            spec_dicts[idx][spec_type]
        except IndexError:
            print(f"Unable to access requested spec_dict (idx {idx})")
        except KeyError:
            print(f"Unable to access '{spec_type}' for "
                f"requested spec_dict (idx {idx})")
        except Exception as e:
            print(f"Unexpected exception - details: {e}")
        """, is_code=True)
        )

    message = {
        conf.BRIEF: title + unspecific_warning,
        conf.MAIN: title + unspecific_warning + unspecific_demo,
    }
    return message



from superhelp.helpers import filt_block_help
from .. import ast_funcs, conf
from superhelp.gen_utils import layout_comment as layout

FOR_XPATH = 'descendant-or-self::For'
WHILE_XPATH = 'descendant-or-self::While'
IF_XPATH = 'descendant-or-self::If'  ## includes elif
OUTER_XPATHS = {
    'for': FOR_XPATH,
    'while': WHILE_XPATH,
    'if': IF_XPATH,
}
NESTING_XPATH = ' | '.join(OUTER_XPATHS.values())

def _long_nested_block(outer_el):
    _first_line_no, _last_line_no, el_lines_n = ast_funcs.get_el_lines_dets(
        outer_el, ignore_trailing_lines=True)
    nested_block_line_len = el_lines_n - 1  ## ignore the actual outer el e.g. for:
    too_long = nested_block_line_len > conf.MAX_BRIEF_NESTED_BLOCK
    return too_long

def has_long_block(block_el, xpath):
    long_block = False
    outer_els = block_el.xpath(xpath)
    for outer_el in outer_els:
        too_long = _long_nested_block(outer_el)
        if too_long:
            long_block = True
            break
    return long_block

@filt_block_help(xpath=NESTING_XPATH, warning=True)
def bloated_nested_block(block_dets, *, repeat=False):
    """
    Look for long indented blocks under conditionals, inside loops etc that are
    candidates for separating into functions to simplify the narrative of the
    main code.
    """
    bloated_outer_types = set()
    included_if = False
    for lbl, outer_xpath in OUTER_XPATHS.items():
        if has_long_block(block_dets.element, outer_xpath):
            bloated_outer_types.add(lbl)
            if lbl == 'if':
                included_if = True
    if not bloated_outer_types:
        return None

    title = layout("""\
    ### Possibility of avoiding excessively long nested blocks
    """)
    summary_bits = []
    for bloated_outer_type in bloated_outer_types:
        summary_bits.append(layout(f"""\

        The code has at least one long nested block under
        `{bloated_outer_type}:`
        """))
    summary = ''.join(summary_bits)
    short_circuit_msg = layout("""\
    #### Short-circuit and exit early

    It may be possible to unnest the indented code block by exiting early if the
    condition in the `if` expression is not met.
    """)
    short_circuit_demo_msg = (
        layout("""
        For example, instead of:
        """)
        +
        layout("""\
        if tall_enough:
            ## add to basketball team
            line 1
            line 2
            line 3
            ...
            line 30
        logging.info("Finished!")
        """, is_code=True)
        +
        layout("""\
        we could possibly write:
        """)
        +
        layout('''\
        if not tall_enough:
            return
        ## add to basketball team
        line 1
        line 2
        line 3
        ...
        line 30
        logging.info("Finished!")
        ''', is_code=True)
    )
    move_to_func_msg = layout("""\
    #### Shift to function

    It may be possible to pull most of the nested code block into a function
    which can be called instead.
    """)
    move_to_func_demo_msg = (
        layout("""
        For example, instead of:
        """)
        +
        layout("""\
        for name in names:
            ## contact name
            line 1
            line 2
            line 3
            ...
            line 30
        logging.info("Finished!")
        """, is_code=True)
        +
        layout("""\
        we could possibly write:
        """)
        +
        layout('''\
        def contact(name):
            """
            Contact person ...
            """
            line 1
            line 2
            line 3
            ...
            line 30

        for name in names:
            contact(name)
        logging.info("Finished!")
        ''', is_code=True)
    )
    if not repeat:
        brief_strategy = layout("""\
            You might want to consider applying a strategy for avoiding
            excessively long indented blocks:
            """)
        if included_if:
            short_circuit = short_circuit_msg
            short_circuit_demo = short_circuit_demo_msg
        else:
            short_circuit = ''
            short_circuit_demo = ''
        move_to_func = move_to_func_msg
        move_to_func_demo = move_to_func_demo_msg
        human = layout("""\

        Computers can handle lots of nesting without malfunctioning. Human
        brains are not so fortunate. As it says in The Zen of Python:

        > "Flat is better than nested."
        """)
    else:
        brief_strategy = ''
        short_circuit = ''
        short_circuit_demo = ''
        move_to_func = ''
        move_to_func_demo = ''
        human = ''

    message = {
        conf.BRIEF: (title + summary + brief_strategy + short_circuit
            + move_to_func),
        conf.MAIN: (title + summary + brief_strategy + short_circuit
            + short_circuit_demo + move_to_func + move_to_func_demo),
        conf.EXTRA: human,
    }
    return message



from contextlib import contextmanager
from functools import partial
import inspect
import logging
import os
from pathlib import Path
import platform
import re
import sys
import tempfile
from textwrap import dedent, wrap
import webbrowser

import ast

from . import conf
from lxml import etree
import astpath  # @UnresolvedImport

from astpath.asts import _set_encoded_literal, _strip_docstring

## Monkey-patch as at astpath Python 3.8 as at 2020-04-26
## Need to be able to tell val = 1 from val = '1' (that little detail ;-))
## Pull request fixing this was accepted and merged May 2020
def convert_to_xml(node, omit_docstrings=False, node_mappings=None):
    """Convert supplied AST node to XML."""
    possible_docstring = isinstance(
        node, (ast.FunctionDef, ast.ClassDef, ast.Module))

    xml_node = etree.Element(node.__class__.__name__)
    for attr in ('lineno', 'col_offset'):
        value = getattr(node, attr, None)
        if value is not None:
            _set_encoded_literal(
                partial(xml_node.set, attr),
                value
            )
    if node_mappings is not None:
        node_mappings[xml_node] = node

    node_fields = zip(
        node._fields,
        (getattr(node, attr) for attr in node._fields)
    )

    for field_name, field_value in node_fields:
        if isinstance(field_value, ast.AST):
            field = etree.SubElement(xml_node, field_name)
            field.append(
                convert_to_xml(
                    field_value,
                    omit_docstrings,
                    node_mappings,
                )
            )

        elif isinstance(field_value, list):
            field = etree.SubElement(xml_node, field_name)
            if possible_docstring and omit_docstrings and field_name == 'body':
                field_value = _strip_docstring(field_value)

            for item in field_value:
                if isinstance(item, ast.AST):
                    field.append(
                        convert_to_xml(
                            item,
                            omit_docstrings,
                            node_mappings,
                        )
                    )
                else:
                    subfield = etree.SubElement(field, 'item')
                    _set_encoded_literal(
                        partial(setattr, subfield, 'text'),
                        item
                    )

        elif field_value is not None:

            ## The only change is this immediate function call below
            ## add type attribute e.g. so we can distinguish strings from numbers etc
            ## <Constant lineno="1" col_offset="6" type="int" value="1"/>
            _set_encoded_literal(
                partial(xml_node.set, 'type'),
                type(field_value).__name__
            )

            _set_encoded_literal(
                partial(xml_node.set, field_name),
                field_value
            )

    return xml_node

astpath.asts.convert_to_xml = convert_to_xml

starting_num_space_pattern = r"""(?x)
    ^      ## start
    \d{1}  ## one digit
    \s{1}  ## one whitespace
    \S+    ## at least one non-whitespace
    """
starting_num_space_prog = re.compile(starting_num_space_pattern)

def open_output_folder():
    """
    On Linux, webbrowser delegates to xdg-open which checks mime type to decide
    what gets to open something. Directories get delegated by xdg-open to
    nautilus, for example, and html to the webbrowser usually.

    So even though web browsers display the contents of folders very easily we
    can't get xdg-open to delegate a directory to a browser. So I'm cheating - I
    make a wrapper page which has a link to the desired folder uri. Where
    there's a will there's a way.
    """
    project_output_tmpdir = get_superhelp_tmpdir(
        folder=conf.SUPERHELP_PROJECT_OUTPUT)
    project_output_url = project_output_tmpdir.as_uri()
    gen_tmpdir = get_superhelp_tmpdir(folder='superhelp')
    with make_open_tmp_file('project_help.html',
            superhelp_tmpdir=gen_tmpdir, mode='w') as tmp_dets:
        _superhelp_tmpdir, tmp_fh, fpath = tmp_dets
        internal_css = dedent("""\
        body {
          background-color: white;
          margin: 40px 70px 20px 70px;
          max-width: 700px;
        }
        h1, h2 {
          font-weight: bold;
        }
        h1 {
          color: #0072aa;
          font-size: 24px;
        }
        h2 {
          color: black;
          font-size: 18px;
          margin-top: 24px;
        }
        p {
          color: black;
          font-size: 14px;
        }
        svg {
          height: 100px;
          width: 100px;
        }
        a:link, a:visited, a:hover, a:active {
          color: #0072aa;
          border-bottom: 1px solid #0072aa;
          text-decoration: none;
        }
        """)
        head = conf.HTML_HEAD % {'internal_css': internal_css}
        body_inner = f"""\
        <a href='{project_output_url}'>Individual help files for your
        project</a>
        """
        html = f"""\
        <!DOCTYPE html>
        <html lang="en">
        {head}
        <body>
        {conf.LOGO_SVG}
        <h1>SuperHELP for your project</h1>
        <h2>Warning</h2>
        <p>The output for this project is stored in a temporary folder - if you
        want to keep the results you will need to shift
        '{project_output_tmpdir}' somewhere else</p>
        {body_inner}
        </body>
        </html>"""
        tmp_fh.write(html)
    webbrowser.open(fpath.as_uri())

def get_tree(snippet):
    try:
        tree = ast.parse(snippet)
    except SyntaxError as e:
        raise SyntaxError(
            f"Oops - something seems broken in the snippet - details: {e}")
    return tree

def xml_from_tree(tree):
    xml = astpath.asts.convert_to_xml(tree)
    return xml

def get_intro(file_path, *, multi_block=False):
    if not file_path:
        if multi_block:
            intro = (f"Help is provided for the supplied code as a whole and "
                "for each block of code as appropriate. If there is nothing to "
                "say about a block it is skipped.")
        else:
            intro = ''
    else:
        intro = (f"Help is provided for '{file_path}' as a whole and for each "
            "block of code as appropriate. If there is nothing to say about a "
            "block it is skipped.")
    return intro

def get_code_desc(file_path):
    if file_path:
        code_desc = file_path
    else:
        code_desc = 'Overall Snippet'
    return code_desc

def get_line_numbered_snippet(snippet):
    snippet_lines = snippet.split('\n')
    n_lines = len(snippet_lines)
    width = len(str(n_lines))
    new_snippet_lines = []
    for n, line in enumerate(snippet_lines, 1):
        new_snippet_lines.append(f"{n:>{width}}   {line}".rstrip())
    lined_snippet = '\n'.join(new_snippet_lines)
    return lined_snippet

def get_introspected_file_path():
    """
    The actual call we are interested in isn't necessarily the second one (e.g.
    console first then actual script) so we have to explicitly filter for it. In
    pydev, for example, it was the third item.

    https://stackoverflow.com/questions/13699283/how-to-get-the-callers-filename-method-name-in-python
    wasn't correct but gave some hints that I could build upon
    """
    for item in inspect.stack():
        has_superhelp_this = (
            item.code_context is not None
            and 'superhelp.this' in ''.join(item.code_context))  ## seems to be a list of one item in each case
        if has_superhelp_this:
            calling_item = item
            break
    else:  ## didn't break for-loop
        raise Exception('Unable to identify calling script through '
            "introspection. Did you rename 'superhelp' or 'this'? "
            "If that isn't the problem try explicitly supplying "
            "file_path e.g. superhelp.this(file_path=__file__)'")
    file_path = calling_item.filename
    return file_path

def get_os_platform():
    platforms = {
        'Linux': conf.LINUX, 'Windows': conf.WINDOWS, 'Darwin': conf.MAC}
    os_platform = platforms.get(platform.system())
    return os_platform

def clean_path_name(raw_path):
    """
    Get a path we can use in a url or file path.
    """
    clean_path = (raw_path
        .replace('...', '_')
        .replace('/', '_')
        .replace('\\', '_'))
    return clean_path

def get_superhelp_tmpdir(folder='superhelp'):
    tmpdir = tempfile.gettempdir()
    superhelp_tmpdir = os.path.join(tmpdir, folder)
    return Path(superhelp_tmpdir)

@contextmanager
def make_open_tmp_file(fname, *, superhelp_tmpdir=None, mode='w'):
    """
    Note - file needs to be closed for changes to be saved to the file -
    otherwise it will be 0 bytes. Up to client code to ensure it is closed and
    properly available for subsequent steps.
    """
    try:
        if not superhelp_tmpdir:
            superhelp_tmpdir = get_superhelp_tmpdir()
        superhelp_tmpdir.mkdir(exist_ok=True)
        tmp_fh = tempfile.NamedTemporaryFile(
            mode=mode, delete=False, dir=superhelp_tmpdir)
        randomly_named_fpath = Path(tmp_fh.name)
        fpath = Path(randomly_named_fpath.parent) / fname
        os.rename(randomly_named_fpath, fpath)
        yield superhelp_tmpdir, tmp_fh, fpath
    finally:
        tmp_fh.close()

def get_python_version():
    major, minor = sys.version_info[:2]
    return f"{major}.{minor}"

def get_nice_str_list(items, *, item_glue=', ', quoter='`'):
    """
    Get a nice English phrase listing the items.

    :param sequence items: individual items to put into a phrase.
    :param str quoter: default is backtick because it is expected that the most
     common items will be names (variables).
    :return: nice phrase
    :rtype: str
    """
    nice_str_list = item_glue.join(
        [f"{quoter}{item}{quoter}" for item in items[:-1]])
    if nice_str_list:
        nice_str_list += f"{item_glue}and "
    nice_str_list += f"{quoter}{items[-1]}{quoter}"
    return nice_str_list

def _get_quoted_pair(left, right, pair_glue, left_quoter, right_quoter):
    return (
        f"{left_quoter}{left}{left_quoter}"
        f"{pair_glue}"
        f"{right_quoter}{right}{right_quoter}"
    )

def get_nice_pairs(pairs, *, pair_glue=': ', left_quoter='`', right_quoter='`'):
    """
    Get a nice English phrase listing paired items.

    :param sequence pairs: a sequence of two-tuples
    :param str pair_glue: how to join items in pair
     e.g. ': ' --> a: b
     e.g. ' is assigned to ' --> a is assigned to b
    :param str quoter: default is backtick because it is expected that the most
     common items will be names (variables).
    :return: nice phrase
    :rtype: str
    """
    nice_pairs_list = ', '.join(
        _get_quoted_pair(left, right, pair_glue, left_quoter, right_quoter)
        for left, right in pairs[:-1])
    if nice_pairs_list:
        nice_pairs_list += ', and '
    final_left, final_right = pairs[-1]
    nice_pairs_list += _get_quoted_pair(
        final_left, final_right, pair_glue, left_quoter, right_quoter)
    return nice_pairs_list

def int2nice(num):
    """
    :return: nicer version of number ready for use in sentences
    :rtype: str
    """
    nice = {
        0: 'no',
        1: 'one',
        2: 'two',
        3: 'three',
        4: 'four',
        5: 'five',
        6: 'six',
        7: 'seven',
        8: 'eight',
        9: 'nine',
        10: 'ten',
        11: 'eleven',
        12: 'twelve',
    }
    return nice.get(num, num)

def int2first_etc(num):
    """
    :return: first for 1, second for 2 etc
    :rtype: str
    """
    nice = {
        1: 'first',
        2: 'second',
        3: 'third',
        4: 'fourth',
        5: 'fifth',
        6: 'sixth',
        7: 'seventh',
        8: 'eighth',
        9: 'nineth',
        10: 'tenth',
        11: 'eleventh',
        12: 'twelfth',
    }
    return nice.get(num, num)

def layout_comment(raw_comment, *, is_code=False):
    """
    Layout comment ready for MD. Handles dedenting and code.

    Doesn't break up long lines which are titles otherwise the subsequent parts
    won't carry appropriate heading-level formatting.

    :param str raw_comment: Comment ready to be dedented, consolidated, split
     etc. Usually will have new line characters at start and end to ensure MD
     doesn't lump paragraphs together. Extra new lines have no effect in the
     output but might be useful in the source e.g. to allow automatic line
     breaking after editing content.
    :param bool is_code: if True then special code markers are inserted. These
     are replaced by the appropriate code markers in the displayer.
    :rtype: str
    """
    if '`' in raw_comment and is_code:
        logging.debug("Backtick detected in code which is probably a mistake")
    if is_code:
        lines = (
            [conf.PYTHON_CODE_START]
            + dedent(raw_comment).split('\n')
            + [conf.PYTHON_CODE_END]
        )
        indented_lines = [f"{' ' * 4}{line}" for line in lines]
        comment = f'\n'.join(indented_lines) + '\n'  ## new line at end needed otherwise content of next str (if any) becomes part of code highlighting
    else:
        raw_paragraphs = dedent(raw_comment).split('\n\n')  ## we only split paragraphs if two new lines
        new_paragraphs = []
        for raw_paragraph in raw_paragraphs:
            ## replace internal new lines only - we need to preserve the outer ones
            n_start_new_lines = (
                len(raw_paragraph) - len(raw_paragraph.lstrip('\n')))
            n_end_new_lines = (
                len(raw_paragraph) - len(raw_paragraph.rstrip('\n')))
            paragraph = raw_paragraph.strip()
            one_line_paragraph = paragraph.replace('\n', ' ')  ## actually continuations of same line so no need to put on separate lines
            special_line = False
            res = starting_num_space_prog.match(one_line_paragraph)
            if res is not None:
                special_line = True
            if one_line_paragraph.startswith((
                    '#',  ## could be one hash or multiple depending on heading level
                    '* '  ## trying to detect bulleted (unordered) lists
                )):
                special_line = True
            if special_line:
                wrapped_paragraph_lines = [one_line_paragraph, ]
            else:
                wrapped_paragraph_lines = wrap(one_line_paragraph)
            new_paragraph = (
                (n_start_new_lines * '\n\n')
                + '\n'.join(wrapped_paragraph_lines)
                + (n_end_new_lines * '\n\n')
            )
            new_paragraphs.append(new_paragraph)
        comment = '\n' + '\n\n'.join(new_paragraphs)
    return comment

def get_docstring_start(docstring):
    docstring_start = (
        docstring.lstrip('\n').split('\n\n')[0].strip().replace('\n    ', ' '))
    return docstring_start



from collections import defaultdict, namedtuple, Counter

from superhelp.helpers import filt_block_help
from .. import ast_funcs, conf
from superhelp.gen_utils import int2nice, layout_comment as layout

IfDets = namedtuple('IfDetails',
    'multiple_conditions, missing_else, if_clauses')

ELIF = 'elif'
ELSE = 'else'

STRING = 'string'
NUMBER = 'number'

IF_XPATH = 'descendant-or-self::If'

def add_if_details(if_element, if_clauses):
    """
    If under our <If>'s <orelse> we have another <If>, and it is the only child,
    store an 'elif' in if_details and send the <If> under <orelse> through
    again (the AST nests them even though they are siblings in the original code
    syntax); otherwise store an 'else' in if_details and return.

    <If>s have <orelse>s which either have sole <If>s or not. When there is not
    then we have reached else.
    """
    orelse_el = if_element.xpath('orelse')[0]  ## always has one
    orelse_children = orelse_el.getchildren()
    if not orelse_children:
        return  ## merely an If on its own without clauses
    ## else: if or elif: ?
    just_the_if_under_orelse = (
        len(orelse_children) == 1 and orelse_children[0].tag == 'If')
    has_elif = just_the_if_under_orelse
    if has_elif:
        if_clauses.append(ELIF)
        elif_el = orelse_children[0]
        add_if_details(elif_el, if_clauses)
    else:
        if_clauses.append(ELSE)
    return

def get_ifs_details(block_dets):
    """
    There can be multiple if statements in a snippet so we have to handle each
    of them.

    And in AST the following are very similar:

    if foo:
        ...
    else:
        if bar:
            ...

    if foo:
        ...
    elif bar:
        ...

    They both result in:
    <orelse>
        <If ...>

    So need to look under the 'orelse' for anything other than an 'If' to decide
    whether an 'else' or 'elif'.
    """
    ## skip when if __name__ == '__main__'
    if block_dets.block_code_str.startswith("if __name__ == "):
        return []
    raw_if_els = block_dets.element.xpath(IF_XPATH)
    if_elements = []
    for raw_if_el in raw_if_els:
        ## ignore if really an elif
        parent_el = raw_if_el.getparent()
        has_or_else_parent = (parent_el.tag == 'orelse')
        has_siblings = len(parent_el.getchildren()) > 1
        actually_elif = (has_or_else_parent and not has_siblings)
        if not actually_elif:
            if_elements.append(raw_if_el)
    ifs_details = []
    for if_element in if_elements:
        if_clauses = []
        add_if_details(if_element, if_clauses)
        multiple_conditions = bool(if_clauses)
        if multiple_conditions:
            last_sub_if = if_clauses[-1]
            missing_else = (last_sub_if != ELSE)
        else:
            missing_else = False
        ifs_details.append(
            IfDets(multiple_conditions, missing_else, if_clauses))
    return ifs_details

def _get_if_comment(block_dets):
    """
    Have to cope with multiple if statements and make it nice (unnumbered) when
    only one.
    """
    ifs_details = get_ifs_details(block_dets)
    if_comment = ''
    for n, if_details in enumerate(ifs_details, 1):
        counter = '' if len(ifs_details) == 1 else f" {int2nice(n)}"
        if if_details.multiple_conditions:
            n_elifs = if_details.if_clauses.count(ELIF)
            else_comment = (
                " and no `else` clause." if if_details.missing_else
                else " and an `else` clause.")
            elif_rules_comment = (
                " Note - `else` clauses with an `if`, and only an `if`, "
                "underneath count as `elif` clauses not `else` clauses.")
            if_comment += layout(f"""\

            `if` statement{counter} has {int2nice(n_elifs)} `elif` clauses
            {else_comment}{elif_rules_comment}
            """)
        else:
            if_comment += layout(f"""\

            `if` statement{counter} has no extra clauses e.g. `elif`s or an
            `else`.
            """)
    return if_comment

@filt_block_help(xpath=IF_XPATH)
def if_else_overview(block_dets, *, repeat=False):
    """
    Look at conditional statements using if (apart from if __name__ ==
    '__main__").
    """
    if block_dets.block_code_str.startswith("if __name__ == "):
        return None

    title = layout("""
    ### Conditional statement detected
    """)
    if_comment = _get_if_comment(block_dets)
    if not repeat:
        demo = (
            layout("""\

            When using `if`, `elif`, or `else`, the item evaluated can either be
            an expression or a name (variable). Sometimes your code can be more
            readable if you make an intermediate variable and use that instead.
            Shorter, more readable, easier to check the parts. For example:
            """)
            +
            layout("""\
            if length_matters and len(phrase) > conf.MAX_SANE_LENGTH:
                phrase = chop(phrase)
            """, is_code=True)
            +
            layout("""\
            is shorter but more complex than:
            """)
            +
            layout("""\
            too_long = (
                length_matters
                and len(phrase) > conf.MAX_SANE_LENGTH
            )
            if too_long:
                phrase = chop(phrase)
            """, is_code=True)
        )
    else:
        demo = ''

    message = {
        conf.BRIEF: title + if_comment,
        conf.MAIN: title + if_comment + demo,
    }
    return message

@filt_block_help(xpath=IF_XPATH, warning=True)
def missing_else(block_dets, *, repeat=False):
    """
    Warn about benefits in many cases of adding else clause if missing.
    """
    ifs_details = get_ifs_details(block_dets)
    ifs_dets_missing_else = [
        if_details for if_details in ifs_details if if_details.missing_else]
    if not ifs_dets_missing_else:
        return None

    title = layout("""\

        ### Possibly better with `else` clause

        """)
    summary_bits = []
    for i, if_details in enumerate(ifs_dets_missing_else):
        first = (i == 0)
        counter = (
            '' if len(ifs_dets_missing_else) == 1 else f" {int2nice(i + 1)}")
        summary_bits.append(layout(f"""\
        `if` block{counter} has `elif` clauses but lacks an `else` clause.
        """))
        if first and not repeat:
            summary_bits.append(layout(f"""\

            If your `elif` clauses are trying to handle all expected cases it is
            probably best to include an `else` clause as well just in case
            something unexpected happens.

            Note - `else` clauses with an `if` and only the `if` underneath
            count as `elif` clauses.
                """))
    summary = ''.join(summary_bits)
    dets = (
        layout("""\

        You may have left out the `else` because it is currently impossible that
        this branch will ever be called. You know that you can only receive the
        items currently handled by the `elif`s so `else` can logically never be
        called. And that it true - until it isn't later ;-) - for example, if
        the calling code starts supplying more types of whatever is being
        evaluated in the `elif`s. This problem happens surprisingly often and
        can create nasty bugs that are hard to trace. If you add an `else`
        clause that raises an exception you will instantly know if the expected
        conditions for your conditional are breached and exactly what to fix.

        For example:
        """)
        +
        layout("""\
        ## At the point this code is written we absolutely know
        ## there are only two user types: managers and staff

        if user == 'manager':
            create_alert()
            set_up_manager(user)
        elif user == 'staff':
            set_up_staff(user)
        else:
            ## This can logically never happen so
            ## we might be tempted to leave out the else clause
            pass
        """, is_code=True)
        +
        layout("""\

        But then a new user type is created (e.g. admin) and the logic of the
        entire program breaks all over the place with no clarity about the real
        source.

        It would be better to have an `else` clause as follows so you know
        exactly what violated the program's assumptions:
        """)
        +
        layout("""\
        else:
            raise Exception(f"Unexpected user: '{user}'")
        """, is_code=True)
        )

    message = {
        conf.BRIEF: title + summary,
        conf.MAIN: title + summary + dets,
    }
    return message

def get_split_membership_dets(if_el):
    """
    if x == 'a' or x == 'b' or x == 'c':
        print(x)

    If/test/BoolOp/values/Compare
                                 left/Name id 'x'
                                 comparators/Constant s 'a'
                                 comparators/Num n  etc

    Only provide message using content if all items are of the same type and are
    either numbers or strings.
    """
    compare_els = if_el.xpath('test/BoolOp/values/Compare')
    if not compare_els:
        return None
    left_name_comp_vals = defaultdict(list)
    basic_types = set()
    for compare_el in compare_els:
        left_name_els = compare_el.xpath('left/Name')
        if not left_name_els:
            continue
        left_name = left_name_els[0].get('id')
        if not left_name:
            continue
        comparators_els = compare_el.xpath('comparators')
        if not comparators_els:
            continue
        comparators_el = comparators_els[0]
        comparison_els = comparators_el.getchildren()  ## e.g. Strs, Nums, NameConstants etc
        if not comparison_els:
            continue
        comparison_el = comparison_els[0]  ## Str or Num etc
        comp_val = ast_funcs.str_from_el(comparison_el)
        if comp_val is not None:
            basic_types.add(STRING)
            if len(basic_types) > 1:
                return None
        else:
            comp_val = ast_funcs.num_str_from_el(comparison_el)
            if comp_val is not None:
                basic_types.add(NUMBER)
                if len(basic_types) > 1:
                    return None
            else:
                return None  ## Give up - something is not right
        left_name_comp_vals[left_name].append(comp_val)
    c = Counter(left_name_comp_vals.keys())
    comp_var = c.most_common(1)[0][0]
    comp_vals = left_name_comp_vals[comp_var]
    comp_val_strs = []
    basic_type = basic_types.pop()
    for comp_val in comp_vals:
        if basic_type == STRING:
            quoter = "'" if '"' in comp_val else "'"
            comp_val_strs.append(f"{quoter}{comp_val}{quoter}")
        elif basic_type == NUMBER:
            comp_val_strs.append(str(comp_val))
        else:
            raise ValueError(f"Unexpected basic type: '{basic_type}'")
    comp_vals_gp_str = '[' + ', '.join(comp_val_strs) + ']'
    return comp_var, comp_vals_gp_str

@filt_block_help(xpath=IF_XPATH)
def split_group_membership(block_dets, *, repeat=False):
    """
    Explain how to use in group and not in group rather than multiple
    comparisons.

    if x == 'a' or x == 'b' or x == 'c':
        print(x)
    ==>
    if x in ['a', 'b', 'c']:
        print(x)
    """
    if_els = block_dets.element.xpath(IF_XPATH)
    has_split = False
    for if_el in if_els:
        try:
            comp_var, comp_vals_gp_str = get_split_membership_dets(if_el)
        except TypeError:
            continue
        else:
            has_split = True
            break
    if not has_split:
        return None

    summary = layout(f"""\
    ### Possible option of evaluating group membership

    It looks like `{comp_var}` has been in multiple separate comparisons. In
    Python it is possible to do a simple check of group membership instead.
    """)
    if not repeat:
        demo = (
            layout("""\
            For example:
            """)
            +
            layout(f"""\
            if {comp_var} in {comp_vals_gp_str}:
                ...
            """, is_code=True)
        )
        extra_demo = (
            layout("""\
            Or to check if a variable is NOT in a group:
            """)
            +
            layout(f"""\
            if {comp_var} not in {comp_vals_gp_str}:
                ...
            """, is_code=True)
        )
    else:
        demo = ''
        extra_demo = ''

    message = {
        conf.BRIEF: summary + demo,
        conf.MAIN: summary + demo + extra_demo,
    }
    return message

def get_has_explicit_count(if_el):
    compare_els = if_el.xpath('test/Compare')
    if not compare_els:
        return False
    compare_el = compare_els[0]
    func_name_els = compare_el.xpath('left/Call/func/Name')
    if not func_name_els:
        return False
    len_func = (func_name_els[0].get('id') == 'len')
    if not len_func:
        return False
    ops_els = compare_el.xpath('ops')
    if not ops_els:
        return False
    ops_el = ops_els[0]
    operator_els = ops_el.getchildren()
    if not operator_els:
        return False
    operator_type = operator_els[0].tag  ## e.g. Gt
    comparators_els = compare_el.xpath('comparators')
    if not comparators_els:
        return False
    comparator_el = comparators_els[0]
    comparison_els = comparator_el.getchildren()
    if not comparison_els:
        return False
    comparison_el = comparison_els[0]
    n = ast_funcs.num_str_from_el(comparison_el)
    if n is None:
        return False
    explicit_booleans = [
        ('Gt', '0'),
        ('GtE', '1'),
        ('Eq', '0'),
        ('LtE', '0'),
        ('Lt', '1'),
    ]
    has_explicit_boolean = ((operator_type, n) in explicit_booleans)
    return has_explicit_boolean

@filt_block_help(xpath=IF_XPATH)
def implicit_boolean_enough(block_dets, *, repeat=False):
    """
    Look for cases where an implicit boolean comparison is enough.
    """
    if repeat:
        return None
    if_els = block_dets.element.xpath(IF_XPATH)
    implicit_boolean_possible = False
    for if_el in if_els:
        has_explicit_count = get_has_explicit_count(if_el)
        if has_explicit_count:
            implicit_boolean_possible = True
        else:
            continue
    if not implicit_boolean_possible:
        return None
    title = layout("""\
    ### Possible option of using an implicit boolean
    """)
    if not repeat:
        summary = (
            layout("""\

            There is often no need to check a non-zero length explicitly. In
            Python, "", 0, [], {}, (), `None` all evaluate to False when we ask
            if they are `True` or `False`. And they all evaluate to `True` if
            they contain something.

            So if we have a list `my_list` we can replace:
            """)
            +
            layout("""\
            if len(my_list) > 0:
                ...

            """, is_code=True)
            +
            layout("""\
            with:
            """)
            +
            layout("""\
            if my_list:  ## if empty, evaluates to False otherwise True
                ...

            """, is_code=True)
        )
    else:
        summary = layout("""\
        There is often no need to check a non-zero length explicitly.
        """)
    message = {
        conf.BRIEF: title + summary,
    }
    return message

def could_short_circuit(if_el):
    """
    Is there the potential to take advantage of Python's ability to short-
    circuit and collapse a nested IF into its parent as a single expression?

    The if_el must have an IF as its only child AND that nested IF must have an
    ORELSE as a child. But that ORELSE must have no children.

    :param element if_el: If element
    :return: True if the If has the potential to be short-circuited
    :rtype: bool
    """
    body_els = if_el.xpath('body')
    has_one_body = len(body_els) == 1
    if not has_one_body:
        return False
    body_el = body_els[0]
    body_children_els = body_el.getchildren()
    body_one_child = len(body_children_els) == 1
    if not body_one_child:
        return False
    body_child_el = body_children_els[0]
    sole_child_is_if = body_child_el.tag == 'If'
    if not sole_child_is_if:
        return False
    nested_if_el = body_child_el
    nested_if_orelse_els = nested_if_el.xpath('orelse')
    one_orelse = len(nested_if_orelse_els) == 1
    if not one_orelse:
        return False
    nested_if_orelse_el = nested_if_orelse_els[0]
    orelse_has_children = bool(nested_if_orelse_el.getchildren())
    if orelse_has_children:
        return False
    ## we have an IF with one child which is an IF and the nested IF's ORELSE has no children
    return True

@filt_block_help(xpath=IF_XPATH)
def short_circuit(block_dets, *, repeat=False):
    """
    Look for cases where short-circuiting is possible.
    """
    if_els = block_dets.element.xpath(IF_XPATH)
    could_short_circuit_something = False
    for if_el in if_els:
        if could_short_circuit(if_el):
            could_short_circuit_something = True
            break
    if not could_short_circuit_something:
        return None

    summary = layout("""\
    ### Potential to collapse `if`s (possibly relying on short-circuiting)

    This code contains nested `if`s that could potentially be collapsed into one
    single conditional expression.
    """)
    if not repeat:
        how2short_circuit = layout("""\

        If the second `if` can only be run once the first `if` has been
        evaluated as True it is still possible to combine the two expressions -
        as long as the two expressions are combined with `and` AND the test to
        see if the nested expression can be run comes first. We can do this
        because Python supports "short-circuiting" :-).
        """)
        demo = (
            layout("""\
            For example, we can rewrite the following:
            """)
            +
            layout("""\
            if word is not None:
                ## the next expression would raise TypeError: object of type
                ## 'NoneType' has no len() if word was None and it got this far
                if len(word) > 20:
                    print(f"'{word}' is a long word")

            """, is_code=True)
            +
            layout("""\
            as:
            """)
            +
            layout("""\
            ## the second clause is only evaluated if the first evaluates as True
            ## i.e. we have relied on short-circuiting
            if word is not None and len(word) > 20:
                print(f"'{word}' is a long word")

            """, is_code=True)
        )
    else:
        how2short_circuit = ''
        demo = ''

    message = {
        conf.BRIEF: summary + how2short_circuit,
        conf.MAIN: summary + how2short_circuit + demo,
    }
    return message

def could_any_or_all(if_el):
    could_any = False
    could_all = False
    boolop_val_els = if_el.xpath('descendant::BoolOp/values')
    for boolop_val_el in boolop_val_els:
        n_items = len(boolop_val_el.getchildren())
        if n_items < conf.MIN4ANY_OR_ALL:  ## worth doing any or all
            continue
        boolop_el = boolop_val_el.getparent()
        op_els = boolop_el.xpath('op')
        op_el = op_els[0]
        op_type_el = op_el.getchildren()[0]
        if op_type_el.tag == 'Or':
            could_any = True
        elif op_type_el.tag == 'And':
            could_all = True
        if all([could_any, could_all]):
            break
    return could_any, could_all

@filt_block_help(xpath=IF_XPATH)
def any_all(block_dets, *, repeat=False):
    """
    Look for cases where using built-in any or all functions makes sense.
    """
    if_els = block_dets.element.xpath(IF_XPATH)
    could_any_something = False
    could_all_something = False
    for if_el in if_els:
        could_any, could_all = could_any_or_all(if_el)
        if could_any:
            could_any_something = True
        if could_all:
            could_all_something = True
        if all([could_any_something, could_all_something]):  ## LOL - thought I might use 'all' given the context even though only two items
            break
    if not any([could_any_something, could_all_something]):
        return None

    if all([could_any_something, could_all_something]):
        title_content = "Consider using `any` and `all`"
    elif could_any_something:
        title_content = "Consider using `any`"
    elif could_all_something:
        title_content = "Consider using `all`"
    else:
        raise Exception("Unexpected situation with "
            f"could_any_something ({could_any_something}) "
            f"and could_all_something ({could_all_something})")
    title = layout(f"""\
    ### {title_content}
    """)
    if not repeat:
        summary = layout("""\

        Python has built-in `any` and `all` functions that make your code more
        readable when you're evaluating whether all or any of a group of items
        are True (or whether _not_ all or _not_ any are True). The argument has
        to be an iterable e.g. a list.
        """)
        demo = (
            layout("""\
            For example, instead of:
            """)
            +
            layout("""\
            if current or valid or safe or permitted:
                proceed()
            """, is_code=True)
            +
            layout("""\
            you could write the more semantic:
            """)
            +
            layout("""\
            if all([current, valid, safe, permitted]):
                proceed()
            """, is_code=True)
            +
            layout("""\
            or the equivalent, and, arguably, more readable:
            """)
            +
            layout("""\
            conditions = [current, valid, safe, permitted]
            if all(conditions):
                proceed()

            """, is_code=True)
            +
            layout("""\
            Or if you wanted the negation the syntax would be:
            """)
            +
            layout("""\
            if not all(conditions):
                exit()
            """, is_code=True)
            +
            layout("""\
            The `any` function works the same way
            """)
            +
            layout("""\
            flatmate_positives = [
                has_car, has_appliances, has_tv, great_cook, funny]
            if any(flatmate_positives):
                recruit_to_flat()

            """, is_code=True)
        )
        extra = (
            layout("""\

            It is easy to forget that an iterable has to be passed in as the
            argument rather than the individual items. So:
            """)
            +
            layout("""\
            if any(books, magazines, comics, games):  ## FAIL TypeError: any() takes exactly one argument (4 given)
                relax()

            if any([books, magazines, comics, games]):  ## SUCCESS
                relax()
            """, is_code=True)
        )
    else:
        summary = layout("""\

        Python has built-in `any` and `all` functions that can make your code
        more readable.
        """)
        demo = ''
        extra = ''

    message = {
        conf.BRIEF: title + summary,
        conf.MAIN: title + summary + demo,
        conf.EXTRA: extra,
    }
    return message



from html import unescape
import markdown  # @UnresolvedImport
from markdown.extensions import fenced_code  # @UnresolvedImport
from markdown.extensions.tables import TableExtension  # @UnresolvedImport

from . import ansi_printer, cli_utils, tag_formatting

def get_ansi(md):
    """
    Covert markdown string into formatted and code-highlighted terminal code.
    """
    MD = markdown.Markdown(
        tab_length=4,
        extensions=[
            ansi_printer.AnsiPrintExtension(),
            TableExtension(),
            fenced_code.FencedCodeExtension(),
        ],
    )
    MD.convert(md)
    ansi = MD.ansi + '\n'
    ansi = cli_utils.set_hr_widths(ansi) + "\n"
    ## The raw HTML within the source includes fenced code blocks.
    ## Placeholders are numbered like this in the md, we replace back
    ansi_placeholder_tpl = markdown.util.HTML_PLACEHOLDER
    stash = MD.htmlStash
    for i, html_block_str in enumerate(stash.rawHtmlBlocks):
        ansi_placeholder = ansi_placeholder_tpl % i
        raw = unescape(html_block_str)
        if raw.lower().startswith('<br'):
            raw = '\n'
        pre = '<pre><code'
        if raw.startswith(pre):
            _, raw = raw.split(pre, 1)
            raw = raw.split('>', 1)[1].rsplit('</code>', 1)[0]
            raw = tag_formatting.code(raw.strip(), from_fenced_block=True)
        ansi = ansi.replace(ansi_placeholder, raw)
    return ansi

def main(md):
    ansi = get_ansi(md) if md else ''
    return ansi



from .. import ast_funcs, conf
from ..gen_utils import layout_comment as layout
from . import any_block_help

op_name2symbol = {
    'Add': '+',
    'Sub': '-',
    'Mult': '*',
    'Div': '/',
    'Mod': '%',
    'FloorDiv': '//',
    'Pow': '**',
    'BitAnd': '&',
    'BitOr': '|',
    'BitXor': '^',
    'RShift': '>>',
    'LShift': '<<',
}

@any_block_help()
def compound_operator_possible(block_dets, *, repeat=False):
    """
    Look for code like x = x + 1 and suggest the compound operator option.
    """
    block_el = block_dets.element
    assign_els = block_el.xpath('descendant-or-self::Assign')
    missing_compound = False
    for assign_el in assign_els:
        target_name_els = assign_el.xpath('targets/Name')
        if len(target_name_els) != 1:
            continue
        target_name_el = target_name_els[0]
        target_name = target_name_el.get('id')
        next_name_els = assign_el.xpath('value/BinOp/left/Name')
        if len(next_name_els) != 1:
            continue
        next_name = next_name_els[0].get('id')
        same_name = (target_name == next_name)
        if not same_name:
            continue
        else:
            op_els = assign_el.xpath('value/BinOp/op')
            if len(op_els) != 1:
                raise Exception("Binop didn't have exactly one op")
            op_el = op_els[0]
            op_name = op_el.getchildren()[0].tag
            try:
                op_symbol = op_name2symbol[op_name]
            except KeyError:
                continue
            right_els = assign_el.xpath('value/BinOp/right')
            if len(right_els) != 1:
                raise Exception("BinOp didn't have exactly one value  "
                    "even though a BinOp")
            right_el = right_els[0]
            value_els = right_el.getchildren()  ## might be Constant
            if len(value_els) != 1:
                raise Exception("right didn't have exactly one value even "
                    "though a BinOp")
            val_el = value_els[0]
            result = ast_funcs.val_dets(val_el)
            if result is None:
                raise Exception("Unable to get value from right side of BinOp")
            val, needs_quoting = result

            missing_compound = True
            break
    if not missing_compound:
        return None

    title = layout("""\
    ### Compound operator possible
    """)
    val2go = f'"{val}"' if needs_quoting else val
    brief_msg = layout(f"""\

    `{target_name} = {target_name} {op_symbol} {val2go}` could be replaced
    by `{target_name} {op_symbol}= {val2go}`.
    """)
    if not repeat:
        compound_operators = (
            layout("""\
            Python supports the use of compound operators so the following pairs
            are all equivalent:
            """)
            +
            layout("""\
            x = x + 1
            x += 1

            x = x - 1
            x -= 1

            x = x * 5
            x *= 5
            """, is_code=True)
            + layout("""\
            See more details here:
            <https://www.programiz.com/python-programming/operators>

            `+=` and `-+` are probably the most readable but other compound
            operators become more useful when names used are longer. For example
            the compound versions of the following are arguably cleaner:
            """)
            +
            layout("""\
            eisen_value = eisen_value * 5
            eisen_value *= 5

            eisen_value = eisen_value ** 2
            eisen_value **= 2

            eisen_value = eisen_value \ 10
            eisen_value \= 10
            """, is_code=True)
        )
    else:
        compound_operators = ''

    message = {
        conf.BRIEF: title + brief_msg,
        conf.MAIN: title + brief_msg + compound_operators,
    }
    return message




import logging

from . import conf
from superhelp import ast_versioned_funcs as avf
from superhelp.gen_utils import get_python_version, make_open_tmp_file

def get_el_lines_dets(el, *, ignore_trailing_lines=False):
    """
    How long is the snippet of code that completely wraps up this element? If
    there are no "trailing lines" it is simple because the AST includes line
    numbers. We just get the min and max line number values. But if there are,
    these trailing lines are not part of the AST (in the same way that comments
    are ignored). So:

    d = {
        1: 1}

    d = {
        1: 1,
    }

    are equivalent in the AST. And there are only lines 1-2 in the AST in both
    cases.

    If there are subsequent AST-relevant lines of code, we can find the first of
    those lines, subtract 1, and treat that as the end of our code. If client
    code is worried about trailing empty lines it can strip them off itself.
    Blank lines (or other non-code lines e.g. comments) won't have any effect on
    the execution of the code.

    They do affect our judgement of code length though e.g. to see if the
    snippet is too long or short so the ignore_trailing_lines option is
    provided.

    If there are no subsequent AST-relevant lines of code, we are in the dark
    about where exactly the actual code string (assuming we only have the AST to
    work from). In which case we simply add enough lines to cover all reasonable
    cases e.g. if different levels of nested parentheses/brackets are on
    separate lines. E.g.

    a = (
            (
                (
                    (
                        (
                            (
                                (
                                 ...
                                )
                            )
                        )
                    )
                )
            )
        )
    """
    SAFE_EXTRA_LINES = 10
    line_no_strs = set(el.xpath('descendant-or-self::*[@lineno]/@lineno'))
    line_nos = [int(line_no_str) for line_no_str in line_no_strs]
    if not line_nos:
        first_line_no, last_line_no, el_lines_n = None, None, None
    else:
        first_line_no = min(line_nos)
        last_ast_line_no = max(line_nos)
        if ignore_trailing_lines:
            last_line_no = last_ast_line_no
        else:
            module_el = el.xpath('//Module')[0]
            all_line_no_strs = set(
                module_el.xpath('descendant::*[@lineno]/@lineno'))
            all_line_nos = [
                int(line_no_str) for line_no_str in all_line_no_strs]
            subsequent_line_nos = [line_no for line_no in all_line_nos
                if line_no > last_ast_line_no]
            if not subsequent_line_nos:
                last_line_no = last_ast_line_no + SAFE_EXTRA_LINES
            else:
                last_line_no = min(subsequent_line_nos) - 1
        el_lines_n = last_line_no - first_line_no + 1
    return first_line_no, last_line_no, el_lines_n

def store_ast_output(xml):
    with make_open_tmp_file(conf.AST_OUTPUT_XML_FNAME, mode='w') as tmp_dets:
        _superhelp_tmpdir, _tmp_ast_fh, tmp_ast_output_xml_fpath = tmp_dets
        xml.getroottree().write(
            str(tmp_ast_output_xml_fpath), pretty_print=True)
    logging.info("\n\n\n\n\nUpdating AST\n\n\n\n\n")

def get_standardised_el_dict(el):
    """
    If we want to see if a value is the same we need to strip out items that
    only vary by position within the script e.g. lineno or col_offset.

    :param Element el: element we want to standardise e.g. a Constant
    :return: standardised dictionary of element attributes
    :rtype: dict
    """
    std_el_dict = {key: value for key, value in el.items()
        if key not in conf.NON_STD_EL_KEYS}
    return std_el_dict

## when backward compatibility with 3.6 can be dropped use def __getattr__(name):
## https://stackoverflow.com/questions/2447353/getattr-on-a-module
python_version = get_python_version()

if python_version in (conf.PY3_6, conf.PY3_7):

    val_dets = avf.val_dets_3_7

    assigned_num_els_from_block = avf.assigned_num_els_from_block_3_7
    num_str_from_val = avf.num_str_from_val_3_7
    num_str_from_el = avf.num_str_from_el_3_7

    assigned_str_els_from_block = avf.assigned_str_els_from_block_3_7
    str_from_el = avf.str_from_el_3_7
    str_els_from_block = avf.str_els_from_block_3_7

    dict_key_from_subscript = avf.dict_key_from_subscript_3_7

    _get_var_plus_equalled = avf._get_var_plus_equalled_3_7
    _get_var_equal_plussed = avf._get_var_equal_plussed_3_7
    get_danger_status = avf.get_danger_status_3_7

    get_docstring_from_value = avf.get_docstring_from_value_3_7
    get_slice_dets = avf.get_slice_dets_3_7
    get_nt_lbl_flds = avf.get_nt_lbl_flds_3_7
    get_slice_n = avf.get_slice_n_3_7
    get_str_els_being_combined = avf.get_str_els_being_combined_3_7

elif python_version == conf.PY3_8:

    val_dets = avf.val_dets_3_8

    assigned_num_els_from_block = avf.assigned_num_els_from_block_3_8
    num_str_from_val = avf.num_str_from_val_3_8
    num_str_from_el = avf.num_str_from_el_3_8

    assigned_str_els_from_block = avf.assigned_str_els_from_block_3_8
    str_from_el = avf.str_from_el_3_8
    str_els_from_block = avf.str_els_from_block_3_8

    dict_key_from_subscript = avf.dict_key_from_subscript_3_8

    _get_var_plus_equalled = avf._get_var_plus_equalled_3_8
    _get_var_equal_plussed = avf._get_var_equal_plussed_3_8
    get_danger_status = avf.get_danger_status_3_8
    get_docstring_from_value = avf.get_docstring_from_value_3_8
    get_slice_dets = avf.get_slice_dets_3_8
    get_nt_lbl_flds = avf.get_nt_lbl_flds_3_8
    get_slice_n = avf.get_slice_n_3_8
    get_str_els_being_combined = avf.get_str_els_being_combined_3_8

else:
    raise Exception(f"Unexpected Python version {python_version}")
